"""
QA: Make sure values were correctly converted.
"""

import argparse
import logging
import os
import pprint
from pathlib import Path
# Third-party packages
import pandas as pd
# First-party packages
from drapi.code.drapi.drapi import (getTimestamp,
                                    makeDirPath,
                                    successiveParents)
from drapi.code.drapi.qa.columnConversion import spotCheckColumns


# Functions


if __name__ == "__main__":
    # >>> `Argparse` arguments >>>
    parser = argparse.ArgumentParser()

    # Arguments: Main
    parser.add_argument("--Encounter_ID_1",
                        default="visit_occurrence_id",
                        type=str,
                        choices=["visit_occurrence_id"],
                        help="")
    parser.add_argument("--Encounter_ID_2",
                        default="Encounter # (CSN)",
                        type=str,
                        choices=["Encounter # (CSN)"],
                        help="")
    parser.add_argument("--Patient_ID_1",
                        default="person_id",
                        type=str,
                        choices=["person_id"],
                        help="")
    parser.add_argument("--Patient_ID_2",
                        default="Patient Key",
                        type=str,
                        choices=["Patient Key",
                                 "PatientKey"],
                        help="")
    parser.add_argument("--Provider_ID_1",
                        default="provider_id",
                        type=str,
                        choices=["provider_id"],
                        help="")
    parser.add_argument("--Provider_ID_2",
                        default="Provider Key",
                        type=str,
                        choices=["Provider Key"],
                        help="")

    # Arguments: General
    parser.add_argument("--CHUNKSIZE",
                        default=50000,
                        type=int,
                        help="The number of rows to read at a time from the CSV using Pandas `chunksize`")
    parser.add_argument("--MESSAGE_MODULO_CHUNKS",
                        default=50,
                        type=int,
                        help="How often to print a log message, i.e., print a message every x number of chunks, where x is `MESSAGE_MODULO_CHUNKS`")
    parser.add_argument("--MESSAGE_MODULO_FILES",
                        default=100,
                        type=int,
                        help="How often to print a log message, i.e., print a message every x number of chunks, where x is `MESSAGE_MODULO_FILES`")

    # Arguments: Meta-variables
    parser.add_argument("--PROJECT_DIR_DEPTH",
                        default=2,
                        type=int,
                        help="")
    parser.add_argument("--DATA_REQUEST_DIR_DEPTH",
                        default=4,
                        type=int,
                        help="")
    parser.add_argument("--IRB_DIR_DEPTH",
                        default=3,
                        type=int,
                        help="")
    parser.add_argument("--IDR_DATA_REQUEST_DIR_DEPTH",
                        default=6,
                        type=int,
                        help="")
    parser.add_argument("--ROOT_DIRECTORY",
                        default="IRB_DIRECTORY",
                        type=str,
                        choices=["DATA_REQUEST_DIRECTORY",
                                 "IDR_DATA_REQUEST_DIRECTORY",
                                 "IRB_DIRECTORY",
                                 "PROJECT_OR_PORTION_DIRECTORY"],
                        help="")
    parser.add_argument("--LOG_LEVEL",
                        default=10,
                        type=int,
                        help="""Increase output verbosity. See "logging" module's log level for valid values.""")

    # Arguments: SQL connection settings
    parser.add_argument("--SERVER",
                        default="DWSRSRCH01.shands.ufl.edu",
                        type=str,
                        choices=["Acuo03.shands.ufl.edu",
                                 "EDW.shands.ufl.edu",
                                 "DWSRSRCH01.shands.ufl.edu",
                                 "IDR01.shands.ufl.edu",
                                 "RDW.shands.ufl.edu"],
                        help="")
    parser.add_argument("--DATABASE",
                        default="DWS_PROD",
                        type=str,
                        choices=["DWS_NOTES",
                                 "DWS_OMOP_PROD",
                                 "DWS_OMOP",
                                 "DWS_PROD"],  # TODO Add the i2b2 databases... or all the other databases?
                        help="")
    parser.add_argument("--USER_DOMAIN",
                        default="UFAD",
                        type=str,
                        choices=["UFAD"],
                        help="")
    parser.add_argument("--USERNAME",
                        default=os.environ["USER"],
                        type=str,
                        help="")
    parser.add_argument("--USER_ID",
                        default=None,
                        help="")
    parser.add_argument("--USER_PWD",
                        default=None,
                        help="")

    argNamespace = parser.parse_args()

    # Parsed arguments: Main
    encounterID1 = argNamespace.Encounter_ID_1
    patientID1 = argNamespace.Patient_ID_1
    providerID1 = argNamespace.Provider_ID_1
    encounterID2 = argNamespace.Encounter_ID_2
    patientID2 = argNamespace.Patient_ID_2
    providerID2 = argNamespace.Provider_ID_2

    # Parsed arguments: General
    CHUNKSIZE = argNamespace.CHUNKSIZE
    MESSAGE_MODULO_CHUNKS = argNamespace.MESSAGE_MODULO_CHUNKS
    MESSAGE_MODULO_FILES = argNamespace.MESSAGE_MODULO_FILES

    # Parsed arguments: Meta-variables
    PROJECT_DIR_DEPTH = argNamespace.PROJECT_DIR_DEPTH
    DATA_REQUEST_DIR_DEPTH = argNamespace.DATA_REQUEST_DIR_DEPTH
    IRB_DIR_DEPTH = argNamespace.IRB_DIR_DEPTH
    IDR_DATA_REQUEST_DIR_DEPTH = argNamespace.IDR_DATA_REQUEST_DIR_DEPTH

    ROOT_DIRECTORY = argNamespace.ROOT_DIRECTORY
    LOG_LEVEL = argNamespace.LOG_LEVEL

    # Parsed arguments: SQL connection settings
    SERVER = argNamespace.SERVER
    DATABASE = argNamespace.DATABASE
    USER_DOMAIN = argNamespace.USER_DOMAIN
    USERNAME = argNamespace.USERNAME
    USER_ID = argNamespace.USER_ID
    USER_PWD = argNamespace.USER_PWD
    # <<< `Argparse` arguments <<<

    # Variables: Path construction: General
    runTimestamp = getTimestamp()
    thisFilePath = Path(__file__)
    thisFileStem = thisFilePath.stem
    projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
    dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_DIR_DEPTH)
    IRBDir, _ = successiveParents(thisFilePath.absolute(), IRB_DIR_DEPTH)
    IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
    dataDir = projectDir.joinpath("data")
    if dataDir:
        inputDataDir = dataDir.joinpath("input")
        outputDataDir = dataDir.joinpath("output")
        if outputDataDir:
            runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
    logsDir = projectDir.joinpath("logs")
    if logsDir:
        runLogsDir = logsDir.joinpath(thisFileStem)
    sqlDir = projectDir.joinpath("sql")

    if ROOT_DIRECTORY == "PROJECT_OR_PORTION_DIRECTORY":
        rootDirectory = projectDir
    elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
        rootDirectory = dataRequestDir
    elif ROOT_DIRECTORY == "IRB_DIRECTORY":
        rootDirectory = IRBDir
    elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
        rootDirectory = IDRDataRequestDir
    else:
        raise Exception("An unexpected error occurred.")

    # Variables: Path construction: Project-specific
    pass

    # Variables: SQL Parameters
    if USER_ID:
        uid = USER_ID[:]
    else:
        uid = fr"{USER_DOMAIN}\{USERNAME}"
    if USER_PWD:
        userPwd = USER_PWD
    else:
        userPWD = os.environ["HFA_UFADPWD"]
    conStr = f"mssql+pymssql://{uid}:{userPWD}@{SERVER}/{DATABASE}"

    # Variables: Other
    pass

    # Directory creation: General
    makeDirPath(runOutputDir)
    makeDirPath(runLogsDir)

    # Logging block
    logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
    logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

    logger = logging.getLogger(__name__)

    fileHandler = logging.FileHandler(logpath)
    fileHandler.setLevel(9)
    fileHandler.setFormatter(logFormat)

    streamHandler = logging.StreamHandler()
    streamHandler.setLevel(LOG_LEVEL)
    streamHandler.setFormatter(logFormat)

    logger.addHandler(fileHandler)
    logger.addHandler(streamHandler)

    logger.setLevel(9)

    logger.info(f"""Begin running "{thisFilePath}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Script arguments:


    # Arguments
    `CHUNKSIZE`: "{CHUNKSIZE}"
    `MESSAGE_MODULO_CHUNKS`: "{MESSAGE_MODULO_CHUNKS}"
    `MESSAGE_MODULO_FILES`: "{MESSAGE_MODULO_FILES}"

    # Arguments: General
    `PROJECT_DIR_DEPTH`: "{PROJECT_DIR_DEPTH}" ----------> "{projectDir}"
    `IRB_DIR_DEPTH`: "{IRB_DIR_DEPTH}" --------------> "{IRBDir}"
    `IDR_DATA_REQUEST_DIR_DEPTH`: "{IDR_DATA_REQUEST_DIR_DEPTH}" -> "{IDRDataRequestDir}"

    `LOG_LEVEL` = "{LOG_LEVEL}"

    # Arguments: SQL connection settings
    `SERVER` = "{SERVER}"
    `DATABASE` = "{DATABASE}"
    `USER_DOMAIN` = "{USER_DOMAIN}"
    `USERNAME` = "{USERNAME}"
    `USER_ID` = "{USER_ID}"
    `USER_PWD` = censored
    """)
    argList = argNamespace._get_args() + argNamespace._get_kwargs()
    argListString = pprint.pformat(argList)
    logger.info(f"""Script arguments:\n{argListString}""")

    # Begin module body

    TEST_SCRIPT = False

    if not TEST_SCRIPT:
        LIST_1 = ["/data/herman/mnt/ufhsd/SHANDS/SHARE/DSS/IDR Data Requests/ACTIVE RDRs/Liu/IRB202300703/Intermediate Results/OMOP Portion/data/output/identified/2024-01-23 13-54-26/condition_occurrence.csv"] + [path.absolute() for path in Path("/data/herman/mnt/ufhsd/SHANDS/SHARE/DSS/IDR Data Requests/ACTIVE RDRs/Liu/IRB202300703/OMOP_Structured_Data/data/output/identified").iterdir() if path.suffix == ".csv"]
        LIST_2 = [path.absolute() for path in Path("../Concatenated Results/data/output/convertColumnsGeneral/2024-02-23 18-11-43").iterdir()]
    elif TEST_SCRIPT:
        LIST_1 = ["/data/herman/mnt/ufhsd/SHANDS/SHARE/DSS/IDR Data Requests/ACTIVE RDRs/Liu/IRB202300703/OMOP_Structured_Data/data/output/identified/device_exposure.csv"]
        LIST_2 = ["/data/herman/Projects/IRB202300703/Concatenated Results/data/output/convertColumnsGeneral/2024-02-23 18-11-43/device_exposure.csv"]

    listOfFiles1 = sorted([Path(string) for string in LIST_1])
    listOfFiles2 = sorted([Path(string) for string in LIST_2])

    lof1 = [pathObj.name for pathObj in listOfFiles1]
    lof2 = [pathObj.name for pathObj in listOfFiles2]

    dfFiles = pd.DataFrame([lof1, lof2]).T
    dfFiles.columns = ["Group 1 Files", "Group 2 Files"]

    logger.info(f"\n{dfFiles}")

    columnsToCheck1 = {integer: string for integer, string in [(0, encounterID1), (1, patientID1), (2, providerID1)] if not isinstance(string, type(None))}
    columnsToCheck2 = {integer: string for integer, string in [(0, encounterID2), (1, patientID2), (2, providerID2)] if not isinstance(string, type(None))}

    spotCheckColumns(listOfFiles1=listOfFiles1,
                     listofFiles2=listOfFiles2,
                     connectionString=conStr,
                     logger=logger,
                     moduloChunks=MESSAGE_MODULO_CHUNKS,
                     columnsToCheck1=columnsToCheck1,
                     columnsToCheck2=columnsToCheck2)

    # End module body
    logger.info(f"""Finished running "{thisFilePath.absolute().relative_to(rootDirectory)}".""")



"""
Create a map from OneFlorida de-identified IDs to de-identified IDR Patient Keys
"""

import argparse
import json
import logging
import os
import pprint
from pathlib import Path
from typing import (Literal,
                    Union)
# Third-party packages
import pandas as pd
from sqlalchemy import URL
# Local packages
from drapi.code.drapi.drapi import (getTimestamp,
                                    makeDirPath,
                                    successiveParents)
from drapi.code.drapi.constants.phiVariables import VARIABLE_SUFFIXES_BO
# Super-local imports
from drapi.code.drapi.getData.getData import getData


# Functions
def deIdentificationFunction(value: Union[float, int],
                             protocol_id: str,
                             id_suffix: Literal["ACCT",
                                                "ENC",
                                                "LINK",
                                                "LOC",
                                                "NOTE",
                                                "ORD",
                                                "PAT",
                                                "PROV",
                                                "STN"],
                             secret: int) -> str:
    """
    Creates the de-identified ID based on a valid value using an additive mapping.
    """
    newValue = value + secret
    return f"""{protocol_id}_{id_suffix}_{newValue}"""


def deIdentifyByGroups(value,
                       protocol_id: str,
                       id_suffix: str,
                       secret: int) -> Union[pd.NA, str]:
    """
    A wrapper for `deIdentificationFunction` that handles invalid value groups, like NANs and negative numbers.
    """
    if pd.isna(value):
        return value
    elif value <= 0:
        return f"""{protocol_id}_{id_suffix}_0"""
    else:
        return deIdentificationFunction(value=value,
                                        protocol_id=protocol_id,
                                        id_suffix=id_suffix,
                                        secret=secret)


if __name__ == "__main__":
    # >>> `Argparse` arguments >>>
    parser = argparse.ArgumentParser()

    # Arguments: Main: Multiple input option: Dictionary format
    parser.add_argument("--DICTIONARY_OF_ARGUMENTS",
                        type=json.loads,
                        help="""The input must be of this format: {variableName: {"Encryption Type": encryptionType, "Encryption Secret": encryptionSecret}}""")

    # Arguments: Main: Multiple input option: String format
    # NOTE TODO Not implemented yet. This should allow the user to input a space-delimitted list of arguments, easier than the dictionary option
    parser.add_argument("--ARGUMENTS")

    # Arguments: Main: Single input option: Data sources
    parser.add_argument("--MAP_1_PATH",
                        type=str)
    parser.add_argument("--PATIENT_KEYS_FILE_PATH",
                        type=str)
    parser.add_argument("--FIRST_TIME",
                        type=bool)
    parser.add_argument("--OLD_RUN_PATH",
                        type=str,
                        help="The path to the data previously downloaded from another session of this program.")
    parser.add_argument("--MAP_1_COLUMN_NAME_FROM",
                        type=str,
                        help="The column header for the OneFlorida patient ID values.")

    # Arguments: Main: Single input option: De-identification settings
    parser.add_argument("--PROTOCOL_ID",
                        type=str)
    parser.add_argument("--ID_SUFFIX",
                        type=str)
    parser.add_argument("--SECRET",
                        type=int)

    # Arguments: Main
    parser.add_argument("SCRIPT_TEST_MODE",
                        type=lambda stringValue: True if stringValue.lower() == "true" else False if stringValue.lower() == "false" else None,
                        help=""" Choose one of {{True, False}}""")

    # Arguments: General
    parser.add_argument("--CHUNKSIZE",
                        default=50000,
                        type=int,
                        help="The number of rows to read at a time from the CSV using Pandas `chunksize`")
    parser.add_argument("--MESSAGE_MODULO_CHUNKS",
                        default=50,
                        type=int,
                        help="How often to print a log message, i.e., print a message every x number of chunks, where x is `MESSAGE_MODULO_CHUNKS`")
    parser.add_argument("--MESSAGE_MODULO_FILES",
                        default=100,
                        type=int,
                        help="How often to print a log message, i.e., print a message every x number of chunks, where x is `MESSAGE_MODULO_FILES`")

    # Arguments: Meta-parameters
    parser.add_argument("--PROJECT_DIR_DEPTH",
                        default=2,
                        type=int,
                        help="")
    parser.add_argument("--DATA_REQUEST_DIR_DEPTH",
                        default=4,
                        type=int,
                        help="")
    parser.add_argument("--IRB_DIR_DEPTH",
                        default=4,
                        type=int,
                        help="")
    parser.add_argument("--IDR_DATA_REQUEST_DIR_DEPTH",
                        default=6,
                        type=int,
                        help="")
    parser.add_argument("--ROOT_DIRECTORY",
                        default="IRB_DIRECTORY",
                        type=str,
                        choices=["DATA_REQUEST_DIRECTORY",
                                 "IDR_DATA_REQUEST_DIRECTORY",
                                 "IRB_DIRECTORY",
                                 "PROJECT_OR_PORTION_DIRECTORY"],
                        help="")
    parser.add_argument("--LOG_LEVEL",
                        default=10,
                        type=int,
                        help="""Increase output verbosity. See "logging" module's log level for valid values.""")

    # Arguments: SQL connection settings
    parser.add_argument("--SERVER",
                        default="DWSRSRCH01.shands.ufl.edu",
                        type=str,
                        choices=["Acuo03.shands.ufl.edu",
                                 "EDW.shands.ufl.edu",
                                 "DWSRSRCH01.shands.ufl.edu",
                                 "IDR01.shands.ufl.edu",
                                 "RDW.shands.ufl.edu"],
                        help="")
    parser.add_argument("--DATABASE",
                        default="DWS_PROD",
                        type=str,
                        choices=["DWS_NOTES",
                                 "DWS_OMOP_PROD",
                                 "DWS_OMOP",
                                 "DWS_PROD"],  # TODO Add the i2b2 databases... or all the other databases?
                        help="")
    parser.add_argument("--USER_DOMAIN",
                        default="UFAD",
                        type=str,
                        choices=["UFAD"],
                        help="")
    parser.add_argument("--USERNAME",
                        default=os.environ["USER"],
                        type=str,
                        help="")
    parser.add_argument("--USER_ID",
                        default=None,
                        help="")
    parser.add_argument("--USER_PWD",
                        default=None,
                        help="")

    argNamespace = parser.parse_args()

    # Parsed arguments: Main
    DICTIONARY_OF_ARGUMENTS = argNamespace.DICTIONARY_OF_ARGUMENTS
    ARGUMENTS = argNamespace.ARGUMENTS

    MAP_1_PATH = argNamespace.MAP_1_PATH
    PATIENT_KEYS_FILE_PATH = argNamespace.PATIENT_KEYS_FILE_PATH
    FIRST_TIME = argNamespace.FIRST_TIME
    OLD_RUN_PATH = argNamespace.OLD_RUN_PATH
    MAP_1_COLUMN_NAME_FROM = argNamespace.MAP_1_COLUMN_NAME_FROM

    PROTOCOL_ID = argNamespace.PROTOCOL_ID
    SECRET = argNamespace.SECRET
    ID_SUFFIX = argNamespace.ID_SUFFIX

    # >>> meta variable <<<
    SINGLE_VALUE_ARGUMENTS = [MAP_1_PATH,
                              PATIENT_KEYS_FILE_PATH,
                              FIRST_TIME,
                              OLD_RUN_PATH,
                              PROTOCOL_ID,
                              SECRET,
                              ID_SUFFIX]
    # <<< meta variable <<<

    SCRIPT_TEST_MODE = argNamespace.SCRIPT_TEST_MODE

    # Parsed arguments: General
    CHUNKSIZE = argNamespace.CHUNKSIZE
    MESSAGE_MODULO_CHUNKS = argNamespace.MESSAGE_MODULO_CHUNKS
    MESSAGE_MODULO_FILES = argNamespace.MESSAGE_MODULO_FILES

    # Parsed arguments: Meta-parameters
    PROJECT_DIR_DEPTH = argNamespace.PROJECT_DIR_DEPTH
    DATA_REQUEST_DIR_DEPTH = argNamespace.DATA_REQUEST_DIR_DEPTH
    IRB_DIR_DEPTH = argNamespace.IRB_DIR_DEPTH
    IDR_DATA_REQUEST_DIR_DEPTH = argNamespace.IDR_DATA_REQUEST_DIR_DEPTH

    ROOT_DIRECTORY = argNamespace.ROOT_DIRECTORY
    LOG_LEVEL = argNamespace.LOG_LEVEL

    # Parsed arguments: SQL connection settings
    SERVER = argNamespace.SERVER
    DATABASE = argNamespace.DATABASE
    USER_DOMAIN = argNamespace.USER_DOMAIN
    USERNAME = argNamespace.USERNAME
    USER_ID = argNamespace.USER_ID
    USER_PWD = argNamespace.USER_PWD
    # <<< `Argparse` arguments <<<

    # >>> Argument checks >>>
    # NOTE TODO Look into handling this natively with `argparse` by using `subcommands`. See "https://stackoverflow.com/questions/30457162/argparse-with-different-modes"
    if DICTIONARY_OF_ARGUMENTS and any(SINGLE_VALUE_ARGUMENTS):
        message = f"""This program is meant to function one of two ways. Either
1. Pass `DICTIONARY_OF_ARGUMENTS`, or
2. Pass each of
    2. a. `FIRST_TIME`
    2. b. `MAP_1_PATH`
    2. c. `OLD_RUN_PATH`
    2. d. `PATIENT_KEYS_FILE_PATH`
    """
        parser.error(message=message)

    if FIRST_TIME and OLD_RUN_PATH:
        message = "It is ambiguous if you provide both `FIRST_TIME` and `OLD_RUN_PATH`. Please only choose one."
        parser.error(message=message)

    if isinstance(SCRIPT_TEST_MODE, bool):
        pass
    else:
        message = """`SCRIPT_TEST_MODE` Must be one of "True" or "False"."""
        parser.error(message=message)

    # <<< Argument checks <<<

    # Variables: Path construction: General
    runTimestamp = getTimestamp()
    thisFilePath = Path(__file__)
    thisFileStem = thisFilePath.stem
    projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
    dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_DIR_DEPTH)
    IRBDir, _ = successiveParents(thisFilePath.absolute(), IRB_DIR_DEPTH)
    IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
    dataDir = projectDir.joinpath("data")
    if dataDir:
        inputDataDir = dataDir.joinpath("input")
        outputDataDir = dataDir.joinpath("output")
        if outputDataDir:
            runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
    logsDir = projectDir.joinpath("logs")
    if logsDir:
        runLogsDir = logsDir.joinpath(thisFileStem)
    sqlDir = projectDir.joinpath("sql")

    if ROOT_DIRECTORY == "PROJECT_OR_PORTION_DIRECTORY":
        rootDirectory = projectDir
    elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
        rootDirectory = dataRequestDir
    elif ROOT_DIRECTORY == "IRB_DIRECTORY":
        rootDirectory = IRBDir
    elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
        rootDirectory = IDRDataRequestDir
    else:
        raise Exception("An unexpected error occurred.")

    # Variables: Path construction: Project-specific
    pass

    # Variables: SQL Parameters
    if USER_ID:
        userID = USER_ID[:]
    else:
        userID = fr"{USER_DOMAIN}\{USERNAME}"
    if USER_PWD:
        userPwd = USER_PWD
    else:
        userPwd = os.environ["HFA_UFADPWD"]
    conStr = f"mssql+pymssql://{userID}:{userPwd}@{SERVER}/{DATABASE}"

    # Variables: Other
    pass

    # Directory creation: General
    makeDirPath(runOutputDir)
    makeDirPath(runLogsDir)

    # Logging block
    logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
    logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

    logger = logging.getLogger(__name__)

    fileHandler = logging.FileHandler(logpath)
    fileHandler.setLevel(9)
    fileHandler.setFormatter(logFormat)

    streamHandler = logging.StreamHandler()
    streamHandler.setLevel(LOG_LEVEL)
    streamHandler.setFormatter(logFormat)

    logger.addHandler(fileHandler)
    logger.addHandler(streamHandler)

    logger.setLevel(9)

    logger.info(f"""Begin running "{thisFilePath}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Script arguments:

    # Arguments: Meta
    `PROJECT_DIR_DEPTH`: "{PROJECT_DIR_DEPTH}" ----------> "{projectDir}"
    `IRB_DIR_DEPTH`: "{IRB_DIR_DEPTH}" --------------> "{IRBDir}"
    `IDR_DATA_REQUEST_DIR_DEPTH`: "{IDR_DATA_REQUEST_DIR_DEPTH}" -> "{IDRDataRequestDir}"

    """)
    argList = argNamespace._get_args() + argNamespace._get_kwargs()
    argListString = pprint.pformat(argList)  # TODO Remove secrets from list to print, e.g., passwords.
    logger.info(f"""Script arguments:\n{argListString}""")

    # >>> Begin module body >>>

    # Map 1: OneFlorida de-identification map
    map1df = pd.read_csv(MAP_1_PATH)

    # Map 2: OneFlorida to IDR map
    patientKeys = pd.read_csv(PATIENT_KEYS_FILE_PATH)["PatientKey"]
    if FIRST_TIME:
        map2dir = runOutputDir.joinpath("Map 2")
        map2dir.mkdir()
        connectionString = URL.create(drivername="mssql+pymssql",
                                      username=userID,
                                      password=userPwd,
                                      host=SERVER,
                                      database=DATABASE)

        getData(sqlFilePath="../../Intermediate Results/SQL Portion/sql/OneFlorida to IDR Patient Map.SQL",
                connectionString=connectionString,
                filterVariableChunkSize=10000,
                filterVariableColumnName=MAP_1_COLUMN_NAME_FROM,
                filterVariableData=map1df,
                filterVariableFilePath=None,
                filterVariablePythonDataType="int",
                filterVariableSqlQueryTemplatePlaceholder="{ PYTHON_VARIABLE: IDTypeValues }",
                logger=logger,
                outputFileName="OneFlorida to IDR Patient Map",
                runOutputDir=map2dir,
                queryChunkSize=10000)
    else:
        map2dir = Path(OLD_RUN_PATH)

    # Join maps: Concatenate Map 2
    listofPaths = sorted([fpath for fpath in map2dir.iterdir() if fpath.suffix.lower() == ".csv"])
    map2df = pd.DataFrame()
    itTotal = len(listofPaths)
    for it, fpath in enumerate(listofPaths, start=1):
        logger.info(f"""  Working on file {it:,} of {itTotal}.""")
        df = pd.read_csv(fpath)
        map2df = pd.concat([map2df, df])
    del df

    # Map 3: IDR de-identification map
    map3df = pd.DataFrame()
    map3df["Patient Key"] = map2df["Patient Key"]
    map3df["De-identified Patient Key"] = map3df["Patient Key"].apply(lambda value: deIdentifyByGroups(value=value,
                                                                                                       protocol_id=PROTOCOL_ID,
                                                                                                       id_suffix=ID_SUFFIX,
                                                                                                       secret=SECRET))

    # Join maps: Homogenize variable names
    map1df = map1df.rename(columns={"deid_patient_id": "De-identififed OneFlorida Patient ID",
                                    MAP_1_COLUMN_NAME_FROM: "OneFlorida Patient ID"})
    map2df = map2df.rename(columns={"": ""})
    map3df = map3df.rename(columns={"De-identified PatientKey": "Patient Key",
                                    "De-identified PatientKey.1": "De-identified Patient Key"})

    # Join maps
    df = map1df.set_index("OneFlorida Patient ID").join(other=map2df.set_index("OneFlorida Patient ID"),
                                                        how="outer")
    df = df.reset_index()
    joinedMaps = df.set_index("Patient Key").join(other=map3df.set_index("Patient Key"),
                                                  how="outer")
    joinedMaps = joinedMaps.reset_index()

    # Select data to output
    COLUMNS_TO_EXPORT = ["De-identified Patient Key",
                         "De-identififed OneFlorida Patient ID"]
    finalMap = joinedMaps[COLUMNS_TO_EXPORT]
    finalMap = finalMap.sort_values(by=COLUMNS_TO_EXPORT)
    exportPath = runOutputDir.joinpath("OneFlorida to UFHealth Patient ID Map.CSV")
    finalMap.to_csv(exportPath, index=False)

    # QA
    mapSize = finalMap.dropna().shape[0]
    logger.info(f"""Final map shape after dropping any NAs: {mapSize:,}.""")

    # Output location summary
    logger.info(f"""Script output is located in the following directory: "{runOutputDir.absolute().relative_to(rootDirectory)}".""")

    # <<< End module body <<<
    logger.info(f"""Finished running "{thisFilePath.absolute().relative_to(rootDirectory)}".""")



"""
Module to download data using SQL queries.
"""

import logging
from pathlib import Path
from typing import Union
# Third-party packages
import pandas as pd
import sqlalchemy as sa
from sqlalchemy.exc import ResourceClosedError
# First-party packages
from drapi.code.drapi.classes import SecretString
from drapi.code.drapi.getData.getData_inner import getData_inner
from drapi.code.drapi.getData.getData_outer import getData_outer

# Functions


def getData(sqlFilePath: Union[Path, str],
            connectionString1: SecretString,
            filterVariablePythonDataType: str,
            filterVariableSqlQueryTemplatePlaceholder: str,
            logger: logging.Logger,
            runOutputDir: Union[Path, str],
            downloadData: bool,
            outputFileName: str,
            connectionString2: SecretString,
            newSQLTable_Database: str,
            newSQLTable_Name: str,
            newSQLTable_Schema: str,
            filterVariableChunkSize: int = 1000,
            filterVariableColumnName: str = None,
            filterVariableData: Union[pd.DataFrame, pd.Series] = None,
            filterVariableFilePath: Union[Path, str] = None,
            queryChunkSize: int = 10000):
    """
    Efficiently download data based on a SQL query. Can filter query by one variable.
    """
    # >>> Determine if we're filtering >>>
    # Case 1: We have both `filterVariableFilePath` and `filterVariableColumnName` defined
    # Case 2: We have `filterVariableData` defined.
    # Case 3: Neither case 1 nor case 2
    # Determine if we're filtering: Determine filter type
    # Determine if we're filtering: Determine filter type: Case 1
    if isinstance(filterVariableFilePath, (str, Path)) and isinstance(filterVariableColumnName, (str, int)):
        useFilteringCase1 = True
    else:
        useFilteringCase1 = False
    # Determine if we're filtering: Determine filter type: Case 2
    if isinstance(filterVariableData, (pd.Series, pd.DataFrame)):
        useFilteringCase2 = True
    else:
        useFilteringCase2 = False
    # Determine if we're filtering: Determine filter type: Ambiguous case: case 1 and 2
    if useFilteringCase1 and useFilteringCase2:
        message = "Ambiguous argument input. If you want to filter your query by a variable, you must either (1) pass the variable as a Pandas Series or DataFrame or (2) pass the (2.a) path to the file that contains the data and the (2.b.) column name of the variable."
        logger.critical(message)
        raise Exception(message)
    # Determine if we're filtering: Make decision
    if useFilteringCase1:
        useFiltering = True
    elif useFilteringCase2:
        useFiltering = True
    elif (not useFilteringCase1) and (not useFilteringCase2):
        useFiltering = False
    else:
        message = "An unexpected error occurred."
        logger.critical(message)
        raise Exception(message)
    # <<< Determine if we're filtering <<<

    # >>> Drop SQL table, if necessary >>>
    if downloadData:
        pass
    else:
        query = f"""\
DROP TABLE IF EXISTS [{newSQLTable_Schema}].[{newSQLTable_Name}];\
"""
        with sa.engine.create_engine(url=connectionString2).connect() as connection:
            with connection.begin():
                _ = connection.execute(sa.text(query))
                logger.info("If the table exists in the server, it was dropped.")  # TODO Add SQL feedback message
    # <<< Drop SQL table, if necessary <<<

    # Choose core function
    if not useFiltering:
        getData_inner(connectionString1=connectionString1,
                      logger=logger,
                      queryChunkSize=queryChunkSize,
                      runOutputDir=runOutputDir,
                      sqlQuery=None,
                      downloadData=downloadData,
                      connectionString2=connectionString2,
                      newSQLTable_Name=newSQLTable_Name,
                      newSQLTable_Schema=newSQLTable_Schema,
                      outputName=outputFileName,
                      itstring1=None,
                      numChunks1=None,
                      sqlFilePath=sqlFilePath)
    elif useFiltering:
        getData_outer(connectionString1=connectionString1,
                      filterVariableColumnName=filterVariableColumnName,
                      filterVariableChunkSize=filterVariableChunkSize,
                      filterVariableData=filterVariableData,
                      filterVariableFilePath=filterVariableFilePath,
                      filterVariablePythonDataType=filterVariablePythonDataType,
                      filterVariableSqlQueryTemplatePlaceholder=filterVariableSqlQueryTemplatePlaceholder,
                      logger=logger,
                      outputName=outputFileName,
                      queryChunkSize=queryChunkSize,
                      sqlFilePath=sqlFilePath,
                      downloadData=downloadData,
                      connectionString2=connectionString2,
                      newSQLTable_Database=newSQLTable_Database,
                      newSQLTable_Name=newSQLTable_Name,
                      newSQLTable_Schema=newSQLTable_Schema,
                      runOutputDir=runOutputDir)
    else:
        message = "An unexpected error occurred."
        logger.critical(message)
        raise Exception(message)
    return



"""
Script for downloading data using SQL queries.
"""

import argparse
import json
import logging
import os
import pprint
import shutil
from pathlib import Path
# Third-party packages
import sqlalchemy as sa
# First-party packages
from drapi import __version__ as drapiVersion
from drapi import loggingChoices
from drapi.code.drapi.classes import (SecretString)
from drapi.code.drapi.cli_parsers import parse_string_to_boolean
from drapi.code.drapi.drapi import (choosePathToLog,
                                    getTimestamp,
                                    loggingChoiceParser,
                                    makeDirPath)
from drapi.code.drapi.getData.getData import getData


if __name__ == "__main__":
    # >>> `Argparse` arguments >>>
    parser = argparse.ArgumentParser()

    # Arguments: Multiple query option: Main
    parser.add_argument("--DICTIONARY_OF_ARGUMENTS",
                        type=json.loads)

    # Arguments: Single query option: Main
    parser.add_argument("--CONNECTION_STRING",
                        type=SecretString,
                        help=""".""")
    parser.add_argument("--SQL_FILE_PATH",
                        type=str)
    parser.add_argument("--DOWNLOAD_DATA",
                        required=True,
                        type=parse_string_to_boolean)
    parser.add_argument("--FILTER_VARIABLE_FILE_PATH",
                        type=str)
    parser.add_argument("--FILTER_VARIABLE_COLUMN_NAME",
                        type=str)
    parser.add_argument("--FILTER_VARIABLE_DATA",
                        default=None,
                        choices=[None],
                        help="This has not been implemented. For now, please just use the default value of `None` and instead pass the path to the file with the data and its column header.")
    parser.add_argument("--FILTER_VARIABLE_SQL_QUERY_TEMPLATE_PLACEHOLDER",
                        type=str)
    parser.add_argument("--FILTER_VARIABLE_PYTHON_DATA_TYPE",
                        type=str,
                        choices=["int", "str"],
                        help="""This affects how the data is sorted and if it's quoted or not in the SQL query.""")
    parser.add_argument("--FILTER_VARIABLE_CHUNK_SIZE",
                        default=10000,
                        type=int)
    parser.add_argument("--QUERY_CHUNK_SIZE",
                        default=10000,
                        type=int)

    # Arguments: Single query option: non-download method
    parser.add_argument("--CONNECTION_STRING_2",
                        type=SecretString,
                        help="The connection string to use to upload the resulting data.")
    parser.add_argument("--newSQLTable_Database",
                        choices=["DWS_OMOP"],
                        help="The database of the new table to create in the SQL server.")
    parser.add_argument("--newSQLTable_Name",
                        help="The name of the new table to create in the SQL server.")
    parser.add_argument("--newSQLTable_Schema",
                        choices=["dbo"],
                        help="The schema of the new table to create in the SQL server.")

    # Arguments: Single query option: download data method
    parser.add_argument("--OUTPUT_FILE_NAME",
                        type=str,
                        help="The name to give to the files downloaded.")

    # Arguments: Meta-parameters
    parser.add_argument("--TIMESTAMP",
                        type=str)
    parser.add_argument("--LOG_LEVEL",
                        default=10,
                        type=loggingChoiceParser,
                        choices=loggingChoices,
                        help="""Increase output verbosity. See "logging" module's log level for valid values.""")

    # Arguments: SQL connection settings
    parser.add_argument("--SERVER",
                        default="DWSRSRCH01.shands.ufl.edu",
                        type=str,
                        choices=["Acuo03.shands.ufl.edu",
                                 "EDW.shands.ufl.edu",
                                 "DWSRSRCH01.shands.ufl.edu",
                                 "IDR01.shands.ufl.edu",
                                 "RDW.shands.ufl.edu"],
                        help="")
    parser.add_argument("--DATABASE",
                        default="DWS_PROD",
                        type=str,
                        choices=["DWS_NOTES",
                                 "DWS_OMOP_PROD",
                                 "DWS_OMOP",
                                 "DWS_PROD"],  # TODO Add the i2b2 databases... or all the other databases?
                        help="")
    parser.add_argument("--USER_DOMAIN",
                        default="UFAD",
                        type=str,
                        choices=["UFAD"],
                        help="")
    parser.add_argument("--USERNAME",
                        default=os.environ["USER"],
                        type=str,
                        help="")
    parser.add_argument("--USER_ID",
                        default=None,
                        help="")
    parser.add_argument("--USER_PWD",
                        default=None,
                        type=SecretString,
                        help="")

    argNamespace = parser.parse_args()

    # Parsed arguments: Multiple query option: Main
    DICTIONARY_OF_ARGUMENTS = argNamespace.DICTIONARY_OF_ARGUMENTS

    # Parsed arguments: Single query option: Main
    CONNECTION_STRING = argNamespace.CONNECTION_STRING
    SQL_FILE_PATH = argNamespace.SQL_FILE_PATH
    DOWNLOAD_DATA = argNamespace.DOWNLOAD_DATA
    FILTER_VARIABLE_FILE_PATH = argNamespace.FILTER_VARIABLE_FILE_PATH
    FILTER_VARIABLE_COLUMN_NAME = argNamespace.FILTER_VARIABLE_COLUMN_NAME
    FILTER_VARIABLE_DATA = argNamespace.FILTER_VARIABLE_DATA
    FILTER_VARIABLE_SQL_QUERY_TEMPLATE_PLACEHOLDER = argNamespace.FILTER_VARIABLE_SQL_QUERY_TEMPLATE_PLACEHOLDER
    FILTER_VARIABLE_PYTHON_DATA_TYPE = argNamespace.FILTER_VARIABLE_PYTHON_DATA_TYPE
    FILTER_VARIABLE_CHUNK_SIZE = argNamespace.FILTER_VARIABLE_CHUNK_SIZE
    QUERY_CHUNK_SIZE = argNamespace.QUERY_CHUNK_SIZE

    # Parsed arguments: Single query option: non-download method
    CONNECTION_STRING_2 = argNamespace.CONNECTION_STRING_2
    newSQLTable_Database = argNamespace.newSQLTable_Database
    newSQLTable_Name = argNamespace.newSQLTable_Name
    newSQLTable_Schema = argNamespace.newSQLTable_Schema

    # Parsed arguments: Single query option: download data method
    OUTPUT_FILE_NAME = argNamespace.OUTPUT_FILE_NAME

    # Parsed arguments: Meta-parameters
    TIMESTAMP = argNamespace.TIMESTAMP
    LOG_LEVEL = argNamespace.LOG_LEVEL
    # <<< `Argparse` arguments <<<

    # >>> Argument parsing: Additional checks >>>
    # NOTE TODO Look into handling this natively with `argparse` by using `subcommands`. See "https://stackoverflow.com/questions/30457162/argparse-with-different-modes"
    messageProgramOptions = """\
This program is meant to function one of two ways. Either
1. Pass `DICTIONARY_OF_ARGUMENTS`, or
2. Pass each of the single-option arguments
    - `CONNECTION_STRING`
    - `SQL_FILE_PATH`
    - `FILTER_VARIABLE_FILE_PATH`
    - `FILTER_VARIABLE_COLUMN_NAME`
    - `FILTER_VARIABLE_SQL_QUERY_TEMPLATE_PLACEHOLDER`
    - `FILTER_VARIABLE_PYTHON_DATA_TYPE`
    - `FILTER_VARIABLE_CHUNK_SIZE`
    - `QUERY_CHUNK_SIZE`
    - `DOWNLOAD_DATA`
    - If downloading data:
        - `OUTPUT_FILE_NAME`
    - Else:
        - `newSQLTable_Database`
        - `newSQLTable_Name`
        - `newSQLTable_Schema`\
"""

    # Argument parsing: Additional checks: Multiple-query vs single-query
    # NOTE That `SINGLE_OPTION_ARGUMENTS_MAIN` doesn't include `FILTER_VARIABLE_DATA`, because for now it's only possible value is `None`.
    SINGLE_OPTION_ARGUMENTS_MAIN = [CONNECTION_STRING,
                                    SQL_FILE_PATH,
                                    DOWNLOAD_DATA,
                                    FILTER_VARIABLE_FILE_PATH,
                                    FILTER_VARIABLE_COLUMN_NAME,
                                    FILTER_VARIABLE_SQL_QUERY_TEMPLATE_PLACEHOLDER,
                                    FILTER_VARIABLE_PYTHON_DATA_TYPE,
                                    FILTER_VARIABLE_CHUNK_SIZE,
                                    QUERY_CHUNK_SIZE,
                                    newSQLTable_Database,
                                    newSQLTable_Name,
                                    newSQLTable_Schema,
                                    OUTPUT_FILE_NAME]
    SINGLE_OPTION_ARGUMENTS_MAIN_0 = [CONNECTION_STRING,
                                      SQL_FILE_PATH,
                                      DOWNLOAD_DATA,
                                      OUTPUT_FILE_NAME]
    SINGLE_OPTION_ARGUMENTS_MAIN_1 = [CONNECTION_STRING_2,
                                      newSQLTable_Database,
                                      newSQLTable_Name,
                                      newSQLTable_Schema]
    SINGLE_OPTION_ARGUMENTS_MAIN_2 = [OUTPUT_FILE_NAME]
    if DICTIONARY_OF_ARGUMENTS and any(SINGLE_OPTION_ARGUMENTS_MAIN):
        message = "You supplied arguments for two conflicting options: multiple-query and single-query options. " + messageProgramOptions
        parser.error(message)

    # Argument parsing: Additional checks: Download vs non-download option
    if any(SINGLE_OPTION_ARGUMENTS_MAIN_1) and any(SINGLE_OPTION_ARGUMENTS_MAIN_2):
        message = "You supplied arguments for two conflicting options: download and non-download options. " + messageProgramOptions
        parser.error(message)
    # >>> Argument parsing: Additional checks >>>

    # Variables: Path construction: General
    if TIMESTAMP: 
        runTimestamp = TIMESTAMP
    else:
        runTimestamp = getTimestamp()
    thisFilePath = Path(__file__)
    thisFileStem = thisFilePath.stem
    currentWorkingDir = Path(os.getcwd()).absolute()
    projectDir = currentWorkingDir
    dataDir = projectDir.joinpath("data")
    if dataDir:
        inputDataDir = dataDir.joinpath("input")
        intermediateDataDir = dataDir.joinpath("intermediate")
        outputDataDir = dataDir.joinpath("output")
        if intermediateDataDir:
            runIntermediateDir = intermediateDataDir.joinpath(thisFileStem, runTimestamp)
        if outputDataDir:
            runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
    logsDir = projectDir.joinpath("logs")
    if logsDir:
        runLogsDir = logsDir.joinpath(thisFileStem)
    sqlDir = projectDir.joinpath("sql")

    # Variables: Path construction: Project-specific
    pass

    # Variables: Other
    pass

    # Directory creation: General
    makeDirPath(runIntermediateDir)
    makeDirPath(runOutputDir)
    makeDirPath(runLogsDir)

    # Logging block
    logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
    logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

    logger = logging.getLogger(__name__)

    fileHandler = logging.FileHandler(logpath)
    fileHandler.setLevel(9)
    fileHandler.setFormatter(logFormat)

    streamHandler = logging.StreamHandler()
    streamHandler.setLevel(LOG_LEVEL)
    streamHandler.setFormatter(logFormat)

    logger.addHandler(fileHandler)
    logger.addHandler(streamHandler)

    logger.setLevel(9)

    logger.info(f"""Begin running "{choosePathToLog(path=thisFilePath, rootPath=projectDir)}".""")
    logger.info(f"""DRAPI-Lemur version is "{drapiVersion}".""")
    logger.info(f"""All other paths will be reported in debugging relative to the current working directory: "{choosePathToLog(path=projectDir, rootPath=projectDir)}".""")

    argList = argNamespace._get_args() + argNamespace._get_kwargs()
    argListString = pprint.pformat(argList)  # TODO Remove secrets from list to print, e.g., passwords.
    logger.info(f"""Script arguments:\n{argListString}""")

    # Begin module body

    # Conform mapping arguments
    if DICTIONARY_OF_ARGUMENTS:
        sqlFileSettings = DICTIONARY_OF_ARGUMENTS
    else:
        sqlFileSettings = {0: {"CONNECTION_STRING": CONNECTION_STRING,
                               "FILTER_VARIABLE_FILE_PATH": FILTER_VARIABLE_FILE_PATH,
                               "FILTER_VARIABLE_COLUMN_NAME": FILTER_VARIABLE_COLUMN_NAME,
                               "FILTER_VARIABLE_DATA": FILTER_VARIABLE_DATA,
                               "FILTER_VARIABLE_SQL_QUERY_TEMPLATE_PLACEHOLDER": FILTER_VARIABLE_SQL_QUERY_TEMPLATE_PLACEHOLDER,
                               "FILTER_VARIABLE_PYTHON_DATA_TYPE": FILTER_VARIABLE_PYTHON_DATA_TYPE,
                               "FILTER_VARIABLE_CHUNK_SIZE": FILTER_VARIABLE_CHUNK_SIZE,
                               "QUERY_CHUNK_SIZE": QUERY_CHUNK_SIZE,
                               "SQL_FILE_PATH": SQL_FILE_PATH,
                               "DOWNLOAD_DATA": DOWNLOAD_DATA,
                               "CONNECTION_STRING_2": CONNECTION_STRING_2,
                               "newSQLTable_Database": newSQLTable_Database,
                               "newSQLTable_Name": newSQLTable_Name,
                               "newSQLTable_Schema": newSQLTable_Schema,
                               "OUTPUT_FILE_NAME": OUTPUT_FILE_NAME}}

    # Iterate over SQL file settings
    logger.info("""Iterating over SQL directory contents.""")
    sqlFiles = sorted([(key, di["SQL_FILE_PATH"]) for key, di in sqlFileSettings.items()], key=lambda tu: tu[1])
    for dataKey, sqlFilePath in sqlFiles:
        logger.info(f"""  Working on data key "{dataKey}" with SQL file path "{sqlFilePath}".""")
        fileSettings = sqlFileSettings[dataKey]
        SQL_FILE_PATH = fileSettings["SQL_FILE_PATH"]
        CONNECTION_STRING = fileSettings["CONNECTION_STRING"]
        FILTER_VARIABLE_CHUNK_SIZE = fileSettings["FILTER_VARIABLE_CHUNK_SIZE"]
        FILTER_VARIABLE_COLUMN_NAME = fileSettings["FILTER_VARIABLE_COLUMN_NAME"]
        FILTER_VARIABLE_DATA = fileSettings["FILTER_VARIABLE_DATA"]
        FILTER_VARIABLE_FILE_PATH = fileSettings["FILTER_VARIABLE_FILE_PATH"]
        FILTER_VARIABLE_PYTHON_DATA_TYPE = fileSettings["FILTER_VARIABLE_PYTHON_DATA_TYPE"]
        FILTER_VARIABLE_SQL_QUERY_TEMPLATE_PLACEHOLDER = fileSettings["FILTER_VARIABLE_SQL_QUERY_TEMPLATE_PLACEHOLDER"]
        QUERY_CHUNK_SIZE = fileSettings["QUERY_CHUNK_SIZE"]
        DOWNLOAD_DATA = fileSettings["DOWNLOAD_DATA"]
        CONNECTION_STRING_2 = fileSettings["CONNECTION_STRING_2"]
        newSQLTable_Database = fileSettings["newSQLTable_Database"]
        newSQLTable_Name = fileSettings["newSQLTable_Name"]
        newSQLTable_Schema = fileSettings["newSQLTable_Schema"]
        OUTPUT_FILE_NAME = fileSettings["OUTPUT_FILE_NAME"]
        getData(sqlFilePath=SQL_FILE_PATH,
                connectionString1=CONNECTION_STRING,
                filterVariableChunkSize=FILTER_VARIABLE_CHUNK_SIZE,
                filterVariableColumnName=FILTER_VARIABLE_COLUMN_NAME,
                filterVariableData=FILTER_VARIABLE_DATA,
                filterVariableFilePath=FILTER_VARIABLE_FILE_PATH,
                filterVariablePythonDataType=FILTER_VARIABLE_PYTHON_DATA_TYPE,
                filterVariableSqlQueryTemplatePlaceholder=FILTER_VARIABLE_SQL_QUERY_TEMPLATE_PLACEHOLDER,
                outputFileName=OUTPUT_FILE_NAME,
                queryChunkSize=QUERY_CHUNK_SIZE,
                downloadData=DOWNLOAD_DATA,
                connectionString2=CONNECTION_STRING_2,
                newSQLTable_Database=newSQLTable_Database,
                newSQLTable_Name=newSQLTable_Name,
                newSQLTable_Schema=newSQLTable_Schema,
                runOutputDir=runOutputDir,
                logger=logger)

    # Output location summary
    logger.info(f"""Script output is located in the following directory: "{choosePathToLog(path=runOutputDir, rootPath=projectDir)}".""")

    # Remove intermediate files, unless running in `DEBUG` mode.
    if logger.getEffectiveLevel() > 10:
        logger.info("Removing intermediate files.")
        shutil.rmtree(runIntermediateDir)
        logger.info("Removing intermediate files - done.")

    # End module body
    logger.info(f"""Finished running "{choosePathToLog(path=runOutputDir, rootPath=projectDir)}".""")



"""
__init__ file for "drapi" package
"""

import logging
from pathlib import Path
from os.path import dirname
PATH = Path(dirname(__file__)).absolute()

__version__ = "1.0.47"

# Logging choices
loggingChoices_Numeric_Min = min(logging.getLevelNamesMapping().values())
loggingChoices_Numeric_Max = max(logging.getLevelNamesMapping().values())
loggingChoices_Numeric = list(range(loggingChoices_Numeric_Min, loggingChoices_Numeric_Max + 1))
loggingChoices_String = list(logging.getLevelNamesMapping().keys())
loggingChoices = loggingChoices_Numeric + loggingChoices_String



"""
Originally from "/Volumes/SHARE/DSS/IDR Data Requests/ACTIVE RDRs/Guo/IRB202201080/code/DataPullProject" but since modified.
"""

import pandas as pd
import numpy as np

# import identified cohort file used for rapid data service


def deidentify_shift(cohort_file, data_release):
    person_id_mapping = cohort_file[['person_id']].drop_duplicates()
    person_id_mapping.reset_index(drop=True, inplace=True)
    occurrence_id_mapping = cohort_file[['visit_occurrence_id']].drop_duplicates()
    occurrence_id_mapping.reset_index(drop=True, inplace=True)
    location_id_mapping = cohort_file[['location_id']].drop_duplicates()
    location_id_mapping.reset_index(drop=True, inplace=True)
    # create a deidentified id for each patient key and keep only needed columns
    person_id_mapping['deid_num'] = person_id_mapping.index + 1
    person_id_mapping['Deidentified_ID'] = person_id_mapping.apply(lambda row: 'Pat_{}'.format(row['deid_num']), axis=1)
    person_id_mapping = person_id_mapping[['person_id', 'Deidentified_ID']]
    # create random number for date shifting between +/- 30
    if ('deidentified' in data_release):
        person_id_mapping['Shift_Num'] = np.random.randint(-30, 30, size=len(person_id_mapping))
        person_id_mapping['Shift_Num'] = pd.to_timedelta(person_id_mapping['Shift_Num'], 'd')
    # create a deidentified id for each occurrence_id and keep only needed columns
    occurrence_id_mapping['deid_num'] = occurrence_id_mapping.index + 1
    occurrence_id_mapping['Deidentified_occurrence_id'] = occurrence_id_mapping.apply(lambda row: 'occurrence_id_{}'.format(row['deid_num']), axis=1)
    occurrence_id_mapping = occurrence_id_mapping[['visit_occurrence_id', 'Deidentified_occurrence_id']]

    location_id_mapping['deid_num'] = location_id_mapping.index + 1
    location_id_mapping['Deidentified_location_id'] = location_id_mapping.apply(lambda row: 'location_id_{}'.format(row['deid_num']), axis=1)
    location_id_mapping = location_id_mapping[['location_id', 'Deidentified_location_id']]
    # output mapping files
    return person_id_mapping, occurrence_id_mapping, location_id_mapping


def shift_person_occurrence(current_file, mapping_file_location, data_release):
    mapping_file = pd.read_csv(mapping_file_location)
    id_mapping = mapping_file[['person_id', 'Deidentified_ID']].drop_duplicates()
    occurrence_mapping = mapping_file[['visit_occurrence_id', 'Deidentified_occurrence_id']].drop_duplicates()
    location_mapping = mapping_file[['location_id', 'Deidentified_location_id']]

    dfs = pd.DataFrame()
    for chunk in pd.read_csv(current_file, chunksize=50000):
        deid1 = pd.DataFrame()
        try:
            deid1 = pd.merge(id_mapping, chunk, on=['person_id'])
            deid1.drop(['person_id'], axis=1, inplace=True)
        except Exception as err:
            _ = err
            deid1 = pd.merge(location_mapping, chunk, on=['location_id'])
            deid1.drop(['location_id'], axis=1, inplace=True)
        try:
            deid1 = pd.merge(occurrence_mapping, deid1, on=['visit_occurrence_id'])
            deid1.drop(['visit_occurrence_id'], axis=1, inplace=True)
        except Exception as err:
            _ = err
            pass
        if (data_release[0] == 'deidentified'):
            date_mapping = mapping_file[['person_id', 'Shift_Num']]
            date_cols = [col for col in deid1.columns if 'date' in col]
            for i in range(len(date_cols)):
                current_column = date_cols[i]
                deid1.loc[:, current_column] = pd.to_datetime(deid1[current_column], errors='ignore')
                date_mapping.loc[:, 'Shift_Num'] = pd.to_timedelta(date_mapping['Shift_Num'])
                deid1.loc[:, current_column] = pd.to_datetime(deid1[current_column], errors='ignore') + pd.to_timedelta(date_mapping['Shift_Num'], 'd')
        dfs = dfs.append(deid1, ignore_index=False)
    return dfs


r'''
#import merged demographics file and merge again with the mapping file to a new file deidentified demographics file
print('de-identify demographics')
demographics=pd.read_csv(r'..\data_sample\demographics_fake.csv')
demo_deid=pd.merge(person_id_mapping,demographics, on='person_id')
demo_deid=pd.merge(occurrence_id_mapping,demo_deid, on='encounter_occurrence_id')
demo_deid2=demo_deid[['Deidentified_ID', 'Deidentified_occurrence_id', 'current_age', 'sex', 'race',
                'ethnicity','language']]
demo_deid2.to_csv(r'..\data_sample\de-identified\demographics_fake.csv',index=False,header=True)

#de-identify labs
print('de-identify labs')
labs=pd.read_csv(r'..\data_sample\labs_fake.csv')
labs_deid=pd.merge(person_id_mapping,labs, on='person_id')
labs_deid=pd.merge(occurrence_id_mapping,labs_deid, on='encounter_occurrence_id')
labs_deid['specimen_taken_datetime'] = pd.to_datetime(labs_deid['specimen_taken_datetime'])
labs_deid['specimen_taken_datetime_shifted'] = pd.to_datetime(labs_deid['specimen_taken_datetime']) + labs_deid['Shift_Num']
labs_deid['inferred_specimen_datetime'] = pd.to_datetime(labs_deid['inferred_specimen_datetime'])
labs_deid['inferred_specimen_datetime'] = pd.to_datetime(labs_deid['inferred_specimen_datetime']) + labs_deid['Shift_Num']
labs_deid2=labs_deid[['Deidentified_ID','Deidentified_occurrence_id','hospital','age_at_encounter','lab_order_description','lab_name','lab_result',
lab_result_numeric','lab_unit','LOINC_code','specimen_taken_datetime_shifted']]
labs_deid2.to_csv(r'..\data_sample\de-identified\labs_fake.csv',index=False,header=True)

print('END')
'''



#!/usr/bin/env python

"""
Gather files.
"""

import argparse
import logging
import os
import pprint
import shutil
from pathlib import Path
# Third-party packages
# First-party packages
from drapi import __version__ as drapiVersion
from drapi import loggingChoices
from drapi.code.drapi.cli_parsers import parse_string_to_boolean
from drapi.code.drapi.drapi import (choosePathToLog,
                                    getTimestamp,
                                    loggingChoiceParser,
                                    makeDirPath)
from drapi.code.drapi.gather_files import gather_files


if __name__ == "__main__":
    # >>> `Argparse` arguments >>>
    parser = argparse.ArgumentParser()

    # Arguments: Required
    parser.add_argument("--destination_folder",
                        required=True,
                        type=Path)

    # Arguments: Add mutually exclusive groups
    group = parser.add_mutually_exclusive_group(required=True)
    # Arguments: Add mutually exclusive groups: Group 1
    group.add_argument("--list_of_directories",
                        type=Path,
                        nargs="+")

    # Arguments: Add mutually exclusive groups: Group 2  # NOTE TODO Not implemented yet
    group.add_argument("--list_of_files",
                        type=Path,
                        action="append",
                        nargs="+")

    # Arguments: Options
    parser.add_argument("--list_of_directories_new_names",
                        type=str,
                        nargs="*")
    parser.add_argument("--list_of_loose_files",
                        type=Path,
                        nargs="*",
                        help="These are files that are gathered in the root directory, and do not require a directory or portion name.")

    parser.add_argument("--overwrite_if_exists_archive",
                        type=parse_string_to_boolean)
    parser.add_argument("--overwrite_if_exists_folder",
                        type=parse_string_to_boolean)
    parser.add_argument("--overwrite_if_exists_file",
                        type=parse_string_to_boolean)

    parser.add_argument("--create_merged_folder",
                        type=parse_string_to_boolean)
    parser.add_argument("--create_compressed_archive",
                        type=parse_string_to_boolean)
    parser.add_argument("--delete_folder_after_archiving",
                        type=parse_string_to_boolean)

    # Arguments: Meta-parameters
    parser.add_argument("--TIMESTAMP",
                        type=str)
    parser.add_argument("--LOG_LEVEL",
                        default=10,
                        type=loggingChoiceParser,
                        choices=loggingChoices,
                        help="""Increase output verbosity. See "logging" module's log level for valid values.""")

    argNamespace = parser.parse_args()

    # Parsed arguments: Required
    destination_folder = argNamespace.destination_folder

    # Parsed arguments: Mode 1
    list_of_files = argNamespace.list_of_files

    # Parsed arguments: Mode 2
    list_of_directories = argNamespace.list_of_directories

    # Parsed arguments: Optional arguments
    list_of_directories_new_names = argNamespace.list_of_directories_new_names
    list_of_loose_files = argNamespace.list_of_loose_files
    overwrite_if_exists_folder = argNamespace.overwrite_if_exists_folder
    overwrite_if_exists_file = argNamespace.overwrite_if_exists_file
    create_merged_folder = argNamespace.create_merged_folder
    create_compressed_archive = argNamespace.create_compressed_archive
    delete_folder_after_archiving = argNamespace.delete_folder_after_archiving

    # Parsed arguments: Meta-parameters
    TIMESTAMP = argNamespace.TIMESTAMP
    LOG_LEVEL = argNamespace.LOG_LEVEL
    # <<< `Argparse` arguments <<<

    # >>> Custom argument parsing >>>
    # >>> Custom argument parsing: Parsing 1 >>>
    if list_of_files:
        parser.error("Not implemented. Please do not use this option.")
    else:
        list_of_files = "Not implemented."
        argNamespace.list_of_files = "Not implemented."
    # <<< Custom argument parsing: Parsing 1 <<<

    # >>> Custom argument parsing: Parsing 2 >>>
    pass
    # <<< Custom argument parsing: Parsing 2 <<<
    # <<< Custom argument parsing <<<

    # >>> Argument checks >>>
    # NOTE TODO Look into handling this natively with `argparse` by using `subcommands`. See "https://stackoverflow.com/questions/30457162/argparse-with-different-modes"
    # >>> Argument checks: Check 1 >>>
    pass
    # <<< Argument checks: Check 1 <<<
    # >>> Argument checks: Check 2 >>>
    pass
    # <<< Argument checks: Check 2 <<<
    # <<< Argument checks <<<

    # Variables: Path construction: General
    if TIMESTAMP:
        runTimestamp = TIMESTAMP
    else:
        runTimestamp = getTimestamp()
    thisFilePath = Path(__file__)
    thisFileStem = thisFilePath.stem
    currentWorkingDir = Path(os.getcwd()).absolute()
    projectDir = currentWorkingDir
    dataDir = projectDir.joinpath("data")
    if dataDir:
        inputDataDir = dataDir.joinpath("input")
        intermediateDataDir = dataDir.joinpath("intermediate")
        outputDataDir = dataDir.joinpath("output")
        if intermediateDataDir:
            runIntermediateDir = intermediateDataDir.joinpath(thisFileStem, runTimestamp)
        if outputDataDir:
            runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
    logsDir = projectDir.joinpath("logs")
    if logsDir:
        runLogsDir = logsDir.joinpath(thisFileStem)
    sqlDir = projectDir.joinpath("sql")

    # Variables: Path construction: Project-specific
    pass

    # Variables: Other
    pass

    # Directory creation: General
    makeDirPath(runIntermediateDir)
    makeDirPath(runOutputDir)
    makeDirPath(runLogsDir)

    # Logging block
    logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
    logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

    logger = logging.getLogger(__name__)

    fileHandler = logging.FileHandler(logpath)
    fileHandler.setLevel(9)
    fileHandler.setFormatter(logFormat)

    streamHandler = logging.StreamHandler()
    streamHandler.setLevel(LOG_LEVEL)
    streamHandler.setFormatter(logFormat)

    logger.addHandler(fileHandler)
    logger.addHandler(streamHandler)

    logger.setLevel(9)

    logger.info(f"""Begin running "{choosePathToLog(path=thisFilePath, rootPath=projectDir)}".""")
    logger.info(f"""DRAPI-Lemur version is "{drapiVersion}".""")
    logger.info(f"""All other paths will be reported in debugging relative to the current working directory: "{choosePathToLog(path=projectDir, rootPath=projectDir)}".""")

    argList = argNamespace._get_args() + argNamespace._get_kwargs()
    argListString = pprint.pformat(argList)  # TODO Remove secrets from list to print, e.g., passwords.
    logger.info(f"""Script arguments:\n{argListString}""")

    # >>> Begin script body >>>

    gather_files_parameters = ["destination_folder",
                               "list_of_files",
                               "list_of_directories",
                               "list_of_directories_new_names",
                               "list_of_loose_files",
                               "timestamp",
                               "overwrite_if_exists_archive",
                               "overwrite_if_exists_file",
                               "overwrite_if_exists_folder",
                               "create_merged_folder",
                               "create_compressed_archive",
                               "delete_folder_after_archiving",
                               "logger"]
    gather_files_kwargs = {}
    for argTuple in argList:
        keyword, value = argTuple
        condition_1 = keyword in gather_files_parameters
        condition_2 = not isinstance(value, type(None))
        if condition_1 and condition_2:
            gather_files_kwargs[keyword] = value
    gather_files_kwargs["logger"] = logger

    logger.info(f""" >>> Keyword arguments passed to `gather_files` >>>
{pprint.pformat(gather_files_kwargs)}
 <<< Keyword arguments passed to `gather_files` <<<""")

    gather_files(**gather_files_kwargs)

    # <<< End script body <<<

    # Output location summary
    logger.info(f"""Script output is located in the following directory: "{choosePathToLog(path=runOutputDir, rootPath=projectDir)}".""")

    # Remove intermediate files, unless running in `DEBUG` mode.
    if logger.getEffectiveLevel() > 10:
        logger.info("Removing intermediate files.")
        shutil.rmtree(runIntermediateDir)
        logger.info("Removing intermediate files - done.")

    # Script end confirmation
    logger.info(f"""Finished running "{choosePathToLog(path=runOutputDir, rootPath=projectDir)}".""")



"""
Convert OMOP person IDs to IDR patient keys.

# NOTE Does not expect data in nested directories (e.g., subfolders of "free_text"). Therefore it uses "Path.iterdir" instead of "Path.glob('*/**')".
# NOTE Expects all files to be CSV files. This is because it uses "pd.read_csv".
"""

from pathlib import Path
# Third-party packages
import pandas as pd
from pandas.errors import EmptyDataError
# Local packages
from drapi.drapi import getTimestamp, makeDirPath, personIDs2patientKeys


def makePersonIDMap(COLUMNS_TO_CONVERT,
                    SETS_PATH,
                    listOfPortionDirs,
                    LIST_OF_PORTION_CONDITIONS,
                    CHUNK_SIZE,
                    pipelineOutputDir: Path,
                    ROOT_DIRECTORY: str,
                    rootDirectory,
                    logger) -> Path:
    """
    """
    functionName = __name__.split(".")[-1]
    runOutputDir = pipelineOutputDir.joinpath(functionName, getTimestamp())
    runIntermediateDataDir = pipelineOutputDir.joinpath(functionName, getTimestamp(), "temp")
    makeDirPath(runOutputDir)
    makeDirPath(runIntermediateDataDir)
    logger.info(f"""Begin running "{functionName}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Function arguments:

    # Arguments
    ``: "{""}"
    """)

    # Get set of values
    if SETS_PATH:
        logger.info(f"""Using the set of values previously collected from "{SETS_PATH}".""")
    else:
        logger.info("""Getting the set of values for each variable to convert.""")
        columnSetsVarsDi = {columnName: {"fpath": runIntermediateDataDir.joinpath(f"{columnName}.txt"),
                                         "fileMode": "w"} for columnName in COLUMNS_TO_CONVERT}
        for directory, fileConditions in zip(listOfPortionDirs, LIST_OF_PORTION_CONDITIONS):
            # Act on directory
            logger.info(f"""Working on directory "{directory.absolute().relative_to(rootDirectory)}".""")
            for file in directory.iterdir():
                logger.info(f"""  Working on file "{file.absolute().relative_to(rootDirectory)}".""")
                conditions = [condition(file) for condition in fileConditions]
                if all(conditions):
                    # Read file
                    logger.info("""    File has met all conditions for processing.""")
                    numChunks = sum([1 for _ in pd.read_csv(file, chunksize=CHUNK_SIZE)])
                    dfChunks = pd.read_csv(file, chunksize=CHUNK_SIZE)
                    for it, dfChunk in enumerate(dfChunks, start=1):
                        logger.info(f"""  ..  Working on chunk {it} of {numChunks}.""")
                        for columnName in dfChunk.columns:
                            logger.info(f"""  ..    Working on column "{columnName}".""")
                            if columnName in COLUMNS_TO_CONVERT:
                                logger.info("""  ..  ..  Column must be converted. Collecting values.""")
                                valuesSet = sorted(list(set(dfChunk[columnName].dropna().values)))
                                columnSetFpath = columnSetsVarsDi[columnName]["fpath"]
                                columnSetFileMode = columnSetsVarsDi[columnName]["fileMode"]
                                with open(columnSetFpath, columnSetFileMode) as file:
                                    for value in valuesSet:
                                        file.write(str(value))
                                        file.write("\n")
                                columnSetsVarsDi[columnName]["fileMode"] = "a"
                                logger.info(f"""  ..  ..  Values saved to "{columnSetFpath.absolute().relative_to(rootDirectory)}" in the project directory.""")
                else:
                    logger.info("""    This file does not need to be processed.""")

    # Map values
    if SETS_PATH:
        setsPathDir = SETS_PATH
    else:
        setsPathDir = runIntermediateDataDir
    for file in setsPathDir.iterdir():
        columnName = file.stem
        logger.info(f"""  Working on variable "{columnName}" located at "{file.absolute().relative_to(rootDirectory)}".""")
        # Read file
        try:
            df = pd.read_table(file, header=None)
        except EmptyDataError as err:
            _ = err
            df = pd.DataFrame()
        # Assert
        if df.shape[1] == 1:
            # Try to convert to integer-type
            try:
                df.iloc[:, 0] = df.iloc[:, 0].astype(int)
            except ValueError as err:
                _ = err
            # Check length differences
            len0 = len(df)
            values = set(df.iloc[:, 0].values)
            len1 = len(values)
            logger.info(f"""    The length of the ID array was reduced from {len0:,} to {len1:,} when removing duplicates.""")
        elif df.shape[1] == 0:
            pass
        # Map contents
        map_ = personIDs2patientKeys(list(values))
        # Save map
        mapPath = runOutputDir.joinpath(f"{columnName} map.csv")
        map_.to_csv(mapPath, index=False)
        logger.info(f"""    PersonID-to-PatientKey map saved to "{mapPath.absolute().relative_to(rootDirectory)}".""")

    # Clean up
    # TODO If input directory is empty, delete
    # TODO Delete intermediate run directory

    # Output location summary
    logger.info(f"""Script output is located in the following directory: "{runOutputDir.absolute().relative_to(rootDirectory)}".""")

    # End script
    logger.info(f"""Finished running "{functionName}".""")

    return mapPath



"""
Download data for this portion

NOTE: This might be deprecated if `DICTIONARY_OF_ARGUMENTS` is correctly implemented in the command-line script "getData.py"
"""

import logging
import os
from pathlib import Path
# Third-party packages
import pandas as pd
# First-party packages
from drapi.code.drapi.drapi import (getTimestamp,
                                    makeDirPath,
                                    successiveParents)
from drapi.code.drapi.getData.getData import getData

# Arguments
ARGUMENTS = {1: {"SQL File": "1.SQL"},
             2: {"SQL File": "2.SQL"}}

# Arguments: SQL connection settings
SERVER = "DWSRSRCH01.shands.ufl.edu"
DATABASE = "DWS_PROD"
USERDOMAIN = "UFAD"
USERNAME = os.environ["USER"]
PWD = os.environ["HFA_UFADPWD"]
UID = fr"{USERDOMAIN}\{USERNAME}"


# Arguments: Meta-variables
PROJECT_DIR_DEPTH = 2
DATA_REQUEST_DIR_DEPTH = PROJECT_DIR_DEPTH + 3
IRB_DIR_DEPTH = PROJECT_DIR_DEPTH + 1
IDR_DATA_REQUEST_DIR_DEPTH = PROJECT_DIR_DEPTH + 4

ROOT_DIRECTORY = "IRB_DIRECTORY"  # TODO One of the following:
                                                 # ["IDR_DATA_REQUEST_DIRECTORY",    # noqa
                                                 #  "IRB_DIRECTORY",                 # noqa
                                                 #  "DATA_REQUEST_DIRECTORY",        # noqa
                                                 #  "PROJECT_OR_PORTION_DIRECTORY"]  # noqa

LOG_LEVEL = "INFO"

# Variables: Path construction: General
runTimestamp = getTimestamp()
thisFilePath = Path(__file__)
thisFileStem = thisFilePath.stem
projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_DIR_DEPTH)
IRBDir, _ = successiveParents(thisFilePath.absolute(), IRB_DIR_DEPTH)
IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
dataDir = projectDir.joinpath("data")
if dataDir:
    inputDataDir = dataDir.joinpath("input")
    outputDataDir = dataDir.joinpath("output")
    if outputDataDir:
        runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
logsDir = projectDir.joinpath("logs")
if logsDir:
    runLogsDir = logsDir.joinpath(thisFileStem)
sqlDir = projectDir.joinpath("sql")

if ROOT_DIRECTORY == "PROJECT_OR_PORTION_DIRECTORY":
    rootDirectory = projectDir
elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
    rootDirectory = dataRequestDir
elif ROOT_DIRECTORY == "IRB_DIRECTORY":
    rootDirectory = IRBDir
elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
    rootDirectory = IDRDataRequestDir
else:
    raise Exception("An unexpected error occurred.")

# Variables: Path construction: Project-specific
pass

# Variables: Other
pass

# Directory creation: General
makeDirPath(runOutputDir)
makeDirPath(runLogsDir)

# Logging block
logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

logger = logging.getLogger(__name__)

fileHandler = logging.FileHandler(logpath)
fileHandler.setLevel(9)
fileHandler.setFormatter(logFormat)

streamHandler = logging.StreamHandler()
streamHandler.setLevel(LOG_LEVEL)
streamHandler.setFormatter(logFormat)

logger.addHandler(fileHandler)
logger.addHandler(streamHandler)

logger.setLevel(9)

if __name__ == "__main__":
    logger.info(f"""Begin running "{thisFilePath}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Script arguments:


    # Arguments
    ``: "{"..."}"

    # Arguments: General
    `PROJECT_DIR_DEPTH`: "{PROJECT_DIR_DEPTH}" ----------> "{projectDir}"
    `IRB_DIR_DEPTH`: "{IRB_DIR_DEPTH}" --------------> "{IRBDir}"
    `IDR_DATA_REQUEST_DIR_DEPTH`: "{IDR_DATA_REQUEST_DIR_DEPTH}" -> "{IDRDataRequestDir}"

    `LOG_LEVEL` = "{LOG_LEVEL}"
    """)

    # Script
    for _, di in ARGUMENTS.items():
        sqlFilePath0 = di["SQL File"]
        sqlFilePath = Path(sqlFilePath0)
        outputFileName = sqlFilePath.stem
        logger.info(f"""  Working on SQL file "{sqlFilePath.name}".""")
        getData(sqlFilePath=sqlFilePath,
                connectionString=f"mssql+pymssql://{UID}:{PWD}@{SERVER}/{DATABASE}",
                filterVariableChunkSize=10000,
                filterVariableColumnName="Patient Key",
                filterVariableFilePath="asdf.CSV",
                filterVariablePythonDataType="int",
                filterVariableSqlQueryTemplatePlaceholder="{PYTHON_VARIABLE: Patient Key}",
                logger=logger,
                outputFileName=outputFileName,
                runOutputDir=runOutputDir,
                queryChunkSize=10000)


    # Output location summary
    logger.info(f"""Script output is located in the following directory: "{runOutputDir.absolute().relative_to(rootDirectory)}".""")

    # End script
    logger.info(f"""Finished running "{thisFilePath.absolute().relative_to(rootDirectory)}".""")



"""
Variable constants common to this project
"""

__all__ = ["COLUMNS_TO_DE_IDENTIFY",
           "MODIFIED_OMOP_PORTION_DIR_MAC",
           "MODIFIED_OMOP_PORTION_DIR_WIN",
           "NOTES_PORTION_DIR_MAC",
           "NOTES_PORTION_DIR_WIN",
           "OLD_MAPS_DIR_PATH",
           "OMOP_PORTION_DIR_MAC",
           "OMOP_PORTION_DIR_WIN"]

from pathlib import Path
# Local packages
from drapi.constants.phiVariables import LIST_OF_PHI_VARIABLES_BO, LIST_OF_PHI_VARIABLES_NOTES, LIST_OF_PHI_VARIABLES_OMOP, VARIABLE_SUFFIXES_BO, VARIABLE_SUFFIXES_NOTES, VARIABLE_SUFFIXES_OMOP
from drapi.drapi import successiveParents

# Argument meta variables
IRB_NUMBER = None  # TODO
DATA_REQUEST_ROOT_DIRECTORY_DEPTH = 3  # TODO  # NOTE To prevent unexpected results, like moving, writing, or deleting the wrong files, set this to folder that is the immediate parent of concatenated result and the intermediate results folder.

dataRequestRootDirectory, _ = successiveParents(Path(__file__).absolute(), DATA_REQUEST_ROOT_DIRECTORY_DEPTH)
NOTES_ROOT_DIRECTORY = dataRequestRootDirectory.joinpath("Intermediate Results",
                                                         "Notes Portion",
                                                         "data",
                                                         "output")
# Project arguments
COLUMNS_TO_DE_IDENTIFY = LIST_OF_PHI_VARIABLES_BO + LIST_OF_PHI_VARIABLES_NOTES + LIST_OF_PHI_VARIABLES_OMOP

# `VARIABLE_ALIASES` NOTE: Some variable names are not standardized. This argument is used by the de-identification process when looking for the de-identification map. This way several variables can be de-identified with the same map.
VARIABLE_ALIASES = {"csn": "EncounterCSN",  # TODO This is just an example. Add or remove from this dictionary as necessary.
                    "PatientKey": "Patient Key",
                    "patient_key": "Patient Key",
                    "Patnt Key": "Patient Key"}

VARIABLE_SUFFIXES_LIST = [VARIABLE_SUFFIXES_BO, VARIABLE_SUFFIXES_NOTES, VARIABLE_SUFFIXES_OMOP]
VARIABLE_SUFFIXES = dict()
for variableSuffixDict in VARIABLE_SUFFIXES_LIST:
    VARIABLE_SUFFIXES.update(variableSuffixDict)

# Portion directories
BO_PORTION_DIR_MAC = dataRequestRootDirectory.joinpath("Intermediate Results/BO Portion/data/output/getData/...")  # TODO
BO_PORTION_DIR_WIN = dataRequestRootDirectory.joinpath("Intermediate Results/BO Portion/data/output/getData/...")  # TODO

NOTES_PORTION_DIR_MAC = NOTES_ROOT_DIRECTORY.joinpath("free_text")
NOTES_PORTION_DIR_WIN = NOTES_ROOT_DIRECTORY.joinpath(r"free_text")

OMOP_PORTION_DIR_MAC = dataRequestRootDirectory.joinpath("Intermediate Results/OMOP Portion/data/output/...")  # TODO
OMOP_PORTION_DIR_WIN = dataRequestRootDirectory.joinpath(r"Intermediate Results\OMOP Portion\data\output\...")  # TODO

MODIFIED_OMOP_PORTION_DIR_MAC = Path("data/output/convertColumns/...")  # TODO
MODIFIED_OMOP_PORTION_DIR_WIN = Path("data/output/convertColumns/...")  # TODO

ZIP_CODE_PORTION_DIR_MAC = Path("data/output/convertColumns/...")  # TODO
ZIP_CODE_PORTION_DIR_WIN = Path("data/output/convertColumns/...")  # TODO

# File criteria
NOTES_PORTION_FILE_CRITERIA = [lambda pathObj: pathObj.suffix.lower() == ".csv"]
OMOP_PORTION_FILE_CRITERIA = [lambda pathObj: pathObj.suffix.lower() == ".csv"]
BO_PORTION_FILE_CRITERIA = [lambda pathObj: pathObj.suffix.lower() == ".csv"]
ZIP_CODE_PORTION_FILE_CRITERIA = [None]  # TODO

# Maps
OLD_MAPS_DIR_PATH = {"EncounterCSN": [NOTES_ROOT_DIRECTORY.joinpath("mapping/map_encounter.csv")],
                     "LinkageNoteID": [NOTES_ROOT_DIRECTORY.joinpath("mapping/map_note_link.csv")],
                     "NoteKey": [NOTES_ROOT_DIRECTORY.joinpath("mapping/map_note.csv")],
                     "OrderKey": [NOTES_ROOT_DIRECTORY.joinpath("mapping/map_order.csv")],
                     "PatientKey": [NOTES_ROOT_DIRECTORY.joinpath("mapping/map_patient.csv")],
                     "ProviderKey": [NOTES_ROOT_DIRECTORY.joinpath("mapping/map_provider.csv")]}

# Quality assurance
if __name__ == "__main__":
    ALL_VARS = [dataRequestRootDirectory,
                NOTES_ROOT_DIRECTORY,
                NOTES_PORTION_DIR_MAC,
                NOTES_PORTION_DIR_WIN,
                OMOP_PORTION_DIR_MAC,
                OMOP_PORTION_DIR_WIN,
                MODIFIED_OMOP_PORTION_DIR_MAC,
                MODIFIED_OMOP_PORTION_DIR_WIN]

    for li in OLD_MAPS_DIR_PATH.values():
        ALL_VARS.extend(li)

    for path in ALL_VARS:
        print(path.exists())



from distutils.core import setup
from pathlib import Path
from setuptools import find_namespace_packages

SCRIPTS = sorted([str(fpath) for fpath in Path("./src/scripts").glob("**/*") if fpath.is_file()])

INIT_PATH = Path("src").joinpath("drapi",
                                 "__init__.py")
with open(INIT_PATH) as file:
    for line in file:
        if line.startswith("__version__"):
            VERSION = line.split('"')[1]

setup(name='drapi-lemur',
      package_dir={"": "src"},
      include_package_data=True,
      packages=find_namespace_packages(where="src"),
      scripts=SCRIPTS,
      version=VERSION,
      description='Data Request API for the Integrated Data Repository Research Services of University of Florida.',
      long_description="Data Request API for the Integrated Data Repository Research Services of University of Florida.",
      author='Herman Autore',
      author_email='hf.autore+drapi@gmail.com',
      url='https://github.com/ChemGuy88/hermanCode/archive/refs/tags/v1.0.0.tar.gz',
      keywords=['CTSI',
                'Clinical and Translational Science Institute',
                'IDR',
                'Integrated Data Repository',
                'Integrated Data Repository Research Services',
                'ODSRI',
                'Office of Data Science and Research Implementation',
                'Shands',
                'Sloth 🦥',
                'UF',
                'UF Health',
                'UFHealth',
                'University of Florida'])



"""
Converts OneFlorida patient ID to UF medical record number (MRN).
"""

import os
from pathlib import Path
from typing_extensions import Literal
# Third-party packages
import pandas as pd
# Local packages
from drapi.code.drapi.drapi import replace_sql_query, successiveParents
from drapi.code.drapi.oneFlorida import ID_TYPE_DICT


# Arguments: SQL connection settings
SERVER = "DWSRSRCH01.shands.ufl.edu"
DATABASE = "DWS_PROD"
USERDOMAIN = "UFAD"
USERNAME = os.environ["USER"]
PWD = os.environ["HFA_UFADPWD"]

# Variables: Path construction: General
thisFilePath = Path(__file__)
drapiRootDir, _ = successiveParents(thisFilePath, 3)
sqlDir = drapiRootDir.joinpath("sql")

# Variables: More
sqlFilePath1 = sqlDir.joinpath("ConvertBetweenMrnAndOneFloridaPatID.SQL")
sqlFilePath2 = sqlDir.joinpath("MapOneFloridaIDs.SQL")

# Variables: SQL connection settings
uid = fr"{USERDOMAIN}\{USERNAME}"
conStr = f"mssql+pymssql://{uid}:{PWD}@{SERVER}/{DATABASE}"


def OFID2MRN(OFIDseries: pd.Series) -> pd.DataFrame:
    """

    """
    listAsString = ",".join([str(el) for el in OFIDseries.values])

    with open(sqlFilePath1, "r") as file:
        query0 = file.read()

    query = replace_sql_query(query=query0,
                              old="{PYTHON_VARIABLE: ONE_FLORIDA_PATIENT_IDS}",
                              new=listAsString)

    MRNseries = pd.read_sql(query, con=conStr)

    return MRNseries


def mapOneFloridaIDs(IDTypeValues: pd.Series,
                     IDType: Literal["PATID",
                                     "Patient Key",
                                     "MRN (UF)",
                                     "MRN (Jax)",
                                     "MRN (Pathology)"],
                     returnQueryOnly: bool) -> pd.Series:
    """
    Queries the ID map containing all of the following variables, any of which can be used as a query filter using the `IDType` parameter. The values to query by are used in the `IDTypeValues` parameter.
        | Variable Definition           | Standard or Common Column Name    |
        |-------------------------------|-----------------------------------|
        | OneFlorida patient ID         | OneFlorida Patient ID             |
        | IDR patient key               | Patient Key                       |
        | UF Health Gainesvile MRN      | MRN (UF)                          |
        | UF Health Jacksonville MRN    | MRN (Jax)                         |
        | UF Health Pathology MRN       | MRN (Pathology)                   |
    """

    listAsString = ",".join([f"'{el}'" for el in IDTypeValues.iloc[:].sort_values()])

    with open(sqlFilePath2, "r") as file:
        query0 = file.read()

    query = replace_sql_query(query=query0,
                              old="{PYTHON_VARIABLE: IDTypeValues}",
                              new=listAsString)
    IDTypeInput = IDType.lower()
    IDTypeDict = ID_TYPE_DICT.copy()
    IDTypeSQL = IDTypeDict[IDTypeInput]
    query = replace_sql_query(query=query,
                              old="{PYTHON_VARIABLE: IDTypeSQL}",
                              new=IDTypeSQL)

    if returnQueryOnly:
        return query
    else:
        pass

    dfResults = pd.read_sql(query, con=conStr)

    return dfResults



"""
De-identify files

# NOTE Does not expect data in nested directories (e.g., subfolders of "free_text"). Therefore it uses "Path.iterdir" instead of "Path.glob('*/**')".
# TODO Assign portion name to each path (per OS) so that portion files are stored in their respective folders, this prevents file from being overwritten in the unlikely, but possible, case files from different portions have the same name.
"""

import logging
import sys
from pathlib import Path
# Third-party packages
import pandas as pd
# Local packages
from drapi.code.drapi.drapi import (fileName2variableName,
                                    getTimestamp,
                                    makeDirPath,
                                    makeMap,
                                    map2di,
                                    numericOrString2integerOrString,
                                    readDataFile,
                                    successiveParents)
from drapi.code.drapi.constants.phiVariables import (FILE_NAME_TO_VARIABLE_NAME_DICT,
                                          VARIABLE_NAME_TO_FILE_NAME_DICT)
# Local packages: Script parameters: General
from common import (IRB_NUMBER,
                    ALIAS_DATA_TYPES,
                    COLUMNS_TO_DE_IDENTIFY,
                    DATA_REQUEST_ROOT_DIRECTORY_DEPTH,
                    DATA_TYPES_DICT,
                    VARIABLE_ALIASES,
                    VARIABLE_SUFFIXES)
# Local packages: Script parameters: Paths
from common import (BO_PORTION_DIR_MAC, BO_PORTION_DIR_WIN,
                    I2B2_PORTION_DIR_MAC, I2B2_PORTION_DIR_WIN,
                    MODIFIED_OMOP_PORTION_DIR_MAC, MODIFIED_OMOP_PORTION_DIR_WIN,
                    NOTES_PORTION_DIR_MAC, NOTES_PORTION_DIR_WIN,
                    OMOP_PORTION_DIR_MAC, OMOP_PORTION_DIR_WIN)
# Local packages: Script parameters: File criteria
from common import (BO_PORTION_FILE_CRITERIA,
                    I2B2_PORTION_FILE_CRITERIA,
                    NOTES_PORTION_FILE_CRITERIA,
                    OMOP_PORTION_FILE_CRITERIA)

# Arguments
CONCATENATED_MAPS_DIR_PATH = Path(r"..\Concatenated Results\data\output\concatenateMaps\...")  # TODO

MAPS_DIR_PATH = CONCATENATED_MAPS_DIR_PATH

# Arguments: OMOP data set selection
USE_MODIFIED_OMOP_DATA_SET = True

# Arguments: Portion Paths and conditions
if USE_MODIFIED_OMOP_DATA_SET:
    OMOPPortionDirMac = MODIFIED_OMOP_PORTION_DIR_MAC
    OMOPPortionDirWin = MODIFIED_OMOP_PORTION_DIR_WIN
else:
    OMOPPortionDirMac = OMOP_PORTION_DIR_MAC
    OMOPPortionDirWin = OMOP_PORTION_DIR_WIN

MAC_PATHS = [BO_PORTION_DIR_MAC,
             I2B2_PORTION_DIR_MAC,
             NOTES_PORTION_DIR_MAC,
             OMOPPortionDirMac]
WIN_PATHS = [BO_PORTION_DIR_WIN,
             I2B2_PORTION_DIR_WIN,
             NOTES_PORTION_DIR_WIN,
             OMOPPortionDirWin]

LIST_OF_PORTION_CONDITIONS = [BO_PORTION_FILE_CRITERIA,
                              I2B2_PORTION_FILE_CRITERIA,
                              NOTES_PORTION_FILE_CRITERIA,
                              OMOP_PORTION_FILE_CRITERIA]

# Arguments; General
CHUNK_SIZE = 50000

# Arguments: Meta-variables
CONCATENATED_RESULTS_DIRECTORY_DEPTH = DATA_REQUEST_ROOT_DIRECTORY_DEPTH - 1
PROJECT_DIR_DEPTH = CONCATENATED_RESULTS_DIRECTORY_DEPTH  # The concatenation suite of scripts is considered to be the "project".
IRB_DIR_DEPTH = CONCATENATED_RESULTS_DIRECTORY_DEPTH + 2
IDR_DATA_REQUEST_DIR_DEPTH = IRB_DIR_DEPTH + 3

ROOT_DIRECTORY = "DATA_REQUEST_DIRECTORY"  # TODO One of the following:
                                           # ["IDR_DATA_REQUEST_DIRECTORY",      # noqa
                                           #  "IRB_DIRECTORY",                   # noqa
                                           #  "DATA_REQUEST_DIRECTORY",          # noqa
                                           #  "CONCATENATED_RESULTS_DIRECTORY"]  # noqa

LOG_LEVEL = "INFO"

# Variables: Path construction: General
runTimestamp = getTimestamp()
thisFilePath = Path(__file__)
thisFileStem = thisFilePath.stem
projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_ROOT_DIRECTORY_DEPTH)
IRBDir, _ = successiveParents(thisFilePath.absolute(), IRB_DIR_DEPTH)
IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
dataDir = projectDir.joinpath("data")
if dataDir:
    inputDataDir = dataDir.joinpath("input")
    intermediateDataDir = dataDir.joinpath("intermediate")
    outputDataDir = dataDir.joinpath("output")
    if intermediateDataDir:
        runIntermediateDataDir = intermediateDataDir.joinpath(thisFileStem, runTimestamp)
    if outputDataDir:
        runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
logsDir = projectDir.joinpath("logs")
if logsDir:
    runLogsDir = logsDir.joinpath(thisFileStem)
sqlDir = projectDir.joinpath("sql")

if ROOT_DIRECTORY == "CONCATENATED_RESULTS_DIRECTORY":
    rootDirectory = projectDir
elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
    rootDirectory = dataRequestDir
elif ROOT_DIRECTORY == "IRB_DIRECTORY":
    rootDirectory = IRBDir
elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
    rootDirectory = IDRDataRequestDir

# Variables: Path construction: OS-specific
isAccessible = all([path.exists() for path in MAC_PATHS]) or all([path.exists() for path in WIN_PATHS])
if isAccessible:
    # If you have access to either of the below directories, use this block.
    operatingSystem = sys.platform
    if operatingSystem == "darwin":
        listOfPortionDirs = MAC_PATHS[:]
    elif operatingSystem == "win32":
        listOfPortionDirs = WIN_PATHS[:]
    else:
        raise Exception("Unsupported operating system")
else:
    # If the above option doesn't work, manually copy the database to the `input` directory.
    print("Not implemented. Check settings in your script.")
    sys.exit()

# Directory creation: General
makeDirPath(runIntermediateDataDir)
makeDirPath(runOutputDir)
makeDirPath(runLogsDir)

if __name__ == "__main__":
    # Logging block
    logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
    fileHandler = logging.FileHandler(logpath)
    fileHandler.setLevel(LOG_LEVEL)
    streamHandler = logging.StreamHandler()
    streamHandler.setLevel(LOG_LEVEL)

    logging.basicConfig(format="[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s",
                        handlers=[fileHandler, streamHandler],
                        level=LOG_LEVEL)

    logging.info(f"""Begin running "{thisFilePath}".""")
    logging.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")

    # Assert all portions have file criteria accounted for
    if len(listOfPortionDirs) == len(LIST_OF_PORTION_CONDITIONS):
        pass
    else:
        raise Exception("The number of portions does not equal the number of portion conditions.")

    # Load de-identification maps for each variable that needs to be de-identified
    logging.info("""Loading de-identification maps for each variable that needs to be de-identified.""")
    mapsDi = {}
    mapsColumnNames = {}
    variablesCollected = [FILE_NAME_TO_VARIABLE_NAME_DICT[fileName2variableName(fname)] for fname in MAPS_DIR_PATH.iterdir()]
    for varName in variablesCollected:
        logging.info(f"""  Loading map for "{varName}".""")
        varPath = MAPS_DIR_PATH.joinpath(f"{VARIABLE_NAME_TO_FILE_NAME_DICT[varName]} map.csv")
        map_ = pd.read_csv(varPath)
        mapDi = map2di(map_)
        mapsDi[varName] = mapDi
        mapsColumnNames[varName] = map_.columns[-1]
    # Add aliases to `mapsColumnNames`
    for varName in VARIABLE_ALIASES.keys():
        varAlias = VARIABLE_ALIASES[varName]
        map_ = makeMap(IDset=set(),
                       IDName=varName,
                       startFrom=1,
                       irbNumber=IRB_NUMBER,
                       suffix=VARIABLE_SUFFIXES[varAlias]["deIdIDSuffix"],
                       columnSuffix=varName,
                       deIdentificationMapStyle="lemur",
                       logger=logging.getLogger())
        mapsColumnNames[varName] = map_.columns[-1]
    # Add aliases to `DATA_TYPES_DICT`
    DATA_TYPES_DICT.update(ALIAS_DATA_TYPES)

    # De-identify columns
    logging.info("""De-identifying files.""")
    for directory, fileConditions in zip(listOfPortionDirs, LIST_OF_PORTION_CONDITIONS):
        # Act on directory
        logging.info(f"""Working on directory "{directory.absolute().relative_to(rootDirectory)}".""")
        for file in directory.iterdir():
            logging.info(f"""  Working on file "{file.absolute().relative_to(rootDirectory)}".""")
            conditions = [condition(file) for condition in fileConditions]
            if all(conditions):
                # Set file options
                exportPath = runOutputDir.joinpath(file.name)
                fileMode = "w"
                fileHeaders = True
                # Read file
                logging.info("""    File has met all conditions for processing.""")
                logging.info("""  ..  Reading file to count the number of chunks.""")
                numChunks = sum([1 for _ in readDataFile(file, chunksize=CHUNK_SIZE)])
                logging.info(f"""  ..  This file has {numChunks} chunks.""")
                dfChunks = readDataFile(file, chunksize=CHUNK_SIZE)
                for it, dfChunk in enumerate(dfChunks, start=1):
                    dfChunk = pd.DataFrame(dfChunk)
                    # Work on chunk
                    logging.info(f"""  ..  Working on chunk {it} of {numChunks}.""")
                    for columnName in dfChunk.columns:
                        # Work on column
                        logging.info(f"""  ..    Working on column "{columnName}".""")
                        if columnName in COLUMNS_TO_DE_IDENTIFY:  # Keep this reference to `COLUMNS_TO_DE_IDENTIFY` as a way to make sure that all variables were collected during `getIDValues` and the `makeMap` scripts.
                            variableDataType = DATA_TYPES_DICT[columnName]
                            logging.info(f"""  ..  ..  Column must be de-identified. De-identifying values. Values are being treated as the following data type: "{variableDataType}".""")
                            if columnName in VARIABLE_ALIASES.keys():
                                mapsDiLookUpName = VARIABLE_ALIASES[columnName]
                            else:
                                mapsDiLookUpName = columnName
                            # Look up values in de-identification maps according to data type. NOTE We are relying on pandas assigning the correct data type to the look-up values in the de-identification map.
                            if variableDataType.lower() == "numeric":
                                dfChunk[columnName] = dfChunk[columnName].apply(lambda IDNum: mapsDi[mapsDiLookUpName][IDNum] if not pd.isna(IDNum) else IDNum)
                            elif columnName in ["NRAS"]:
                                dfChunk[columnName] = dfChunk[columnName].apply(lambda IDvalue: f"{IRB_NUMBER}_DE-IDENTIFICATION FAILURE" if not pd.isna(IDvalue) else IDvalue)
                                logging.info("""Using work-around for this variable. This variable is assigned the data type of "Numeric_Or_String", but its values were not reliably mapped to de-identified values.""")  # NOTE Hack. This is the same problem as the other variables below. Part of the problem is how pandas reads files.
                            elif columnName in ["F/u Physicians", "Prim  surgeon Code"]:  # NOTE Hack. These variables are categorized as `numeric_or_string`, but they contain numbers with leading zeros that get converted to integers in the current process. Either we convert these variables to `string` or we change the process.
                                dfChunk[columnName] = dfChunk[columnName].apply(lambda IDvalue: mapsDi[mapsDiLookUpName][str(IDvalue)] if not pd.isna(IDvalue) else IDvalue)
                                logging.info("""Using work-around for this variable. This variable is assigned the data type of "Numeric_Or_String", but is being treated as a "String" variable. If continued use of the "String" data type is successful, then the current assignment of "Numeric_Or_String" should be deprecated.""")
                            elif variableDataType.lower() == "string":
                                dfChunk[columnName] = dfChunk[columnName].apply(lambda IDvalue: mapsDi[mapsDiLookUpName][str(IDvalue)] if not pd.isna(IDvalue) else IDvalue)
                            elif variableDataType.lower() == "numeric_or_string":
                                dfChunk[columnName] = dfChunk[columnName].apply(lambda IDvalue: mapsDi[mapsDiLookUpName][str(numericOrString2integerOrString(IDvalue))] if not pd.isna(IDvalue) else IDvalue)
                            else:
                                msg = "The table column is expected to have a data type associated with it."
                                logging.error(msg)
                                raise ValueError(msg)
                            dfChunk = dfChunk.rename(columns={columnName: mapsColumnNames[columnName]})
                    # Save chunk
                    dfChunk.to_csv(exportPath, mode=fileMode, header=fileHeaders, index=False)
                    fileMode = "a"
                    fileHeaders = False
                    logging.info(f"""  ..  Chunk saved to "{exportPath.absolute().relative_to(rootDirectory)}".""")
            else:
                logging.info("""    This file does not need to be processed.""")

    # Output location summary
    logging.info(f"""Script output is located in the following directory: "{runOutputDir.absolute().relative_to(rootDirectory)}".""")

    # End script
    logging.info(f"""Finished running "{thisFilePath.absolute().relative_to(rootDirectory)}".""")



"""
Utility functions used with images data requests.
"""

__all__ = ["getPatientNumber",
           "getSessionNumber",
           "getPatientAndSessionNumbers"]

import re


def getPatientNumber(string):
    """
    Returns the patient number from an image data release folder name.
    """
    pattern = r"^Patient_(?P<patientNumber>\d+)_IRB201802747$"
    obj = re.match(pattern, string)
    if obj:
        groupdict = obj.groupdict()
        patientNumber = groupdict["patientNumber"]
        if len(patientNumber) > 0:
            patientNumber = int(patientNumber)
        else:
            raise Exception("""String is of an unexpected format: "{string}".""")
    else:
        raise Exception("String was of an unexpected format")
    return patientNumber


def getSessionNumber(string):
    """
    Returns the session number from an image data release folder name.
    """
    pattern = r"^Session_(?P<patientNumber>\d+)_IRB201802747[_]{0,1}(?P<sessionNumber>\d*)$"
    obj = re.match(pattern, string)
    if obj:
        groupdict = obj.groupdict()
        sessionNumber = groupdict["sessionNumber"]
        if len(sessionNumber) > 0:
            sessionNumber = int(sessionNumber)
        elif len(sessionNumber) == 0:
            sessionNumber = 0
        else:
            raise Exception("""String is of an unexpected format: "{string}".""")
    else:
        raise Exception(f"""String was of an unexpected format: "{string}".""")
    return sessionNumber


def getPatientAndSessionNumbers(string):
    """
    Returns the patient and session number from an image data release folder name as a 2-tuple.
    """
    pattern = r"^(?:Session|Patient)_(?P<patientNumber>\d+)_IRB201802747[_]{0,1}(?P<sessionNumber>\d*)$"
    obj = re.match(pattern, string)
    if obj:
        groupdict = obj.groupdict()
        patientNumber = groupdict["patientNumber"]
        if len(patientNumber) > 0:
            patientNumber = int(patientNumber)
        else:
            raise Exception("""String is of an unexpected format: "{string}".""")
        sessionNumber = groupdict["sessionNumber"]
        if len(sessionNumber) > 0:
            sessionNumber = int(sessionNumber)
        elif len(sessionNumber) == 0:
            sessionNumber = 0
        else:
            raise Exception("""String is of an unexpected format: "{string}".""")
    else:
        raise Exception(f"""String was of an unexpected format: "{string}".""")
    return patientNumber, sessionNumber



"""
Create a map from one de-identified ID to another de-identified ID.

Consider the table below for an illustration of the meaning of the parameter names.

|  table1 |   | table 2 |   | table 3 |
| c1 | c2 |   | c3 | c4 |   | c5 | c6 |

Where
`c1` is `LEFT_VARIABLE_DEIDENTIFIED_NAME`
`c2` is `LEFT_VARIABLE_NAME_ALIAS_DEIDENTIFICATION_MAP`
`c3` is `LEFT_VARIABLE_NAME_ALIAS_LEFT_TO_RIGHT_MAP`
`c4` is `RIGHT_VARIABLE_NAME_ALIAS_LEFT_TO_RIGHT_MAP`
`c5` is `RIGHT_VARIABLE_NAME_ALIAS_DEIDENTIFICATION_MAP`
`c6` is `RIGHT_VARIABLE_DEIDENTIFIED_NAME`

Note that 

- `c2` and `c3` can be the same, (e.g., "Patient Key")
- `c4` and `c5` can be the same
- `c1` and `c6` can be the same (e.g., "De-identified Patient Key")

In some rare situations all of `c1` through `c6` can also be the same. For example, if the de-identified name is simply the variable name.

NOTE Situations where `c2` and `c3` are the same or `c4` and `c5` are the same might be handled by using `LEFT_VARIABLE_NAME` or `RIGHT_VARIABLE_NAME`, accordingly. For example, if `LEFT_VARIABLE_NAME` is set, but neither of `LEFT_VARIABLE_NAME_ALIAS_DEIDENTIFICATION_MAP` or `LEFT_VARIABLE_NAME_ALIAS_LEFT_TO_RIGHT_MAP` are set, then apply `LEFT_VARIABLE_NAME` to both of `c2` and `c3`.
"""

import argparse
import logging
import os
import pprint
import shutil
from pathlib import Path
# Third-party packages
import pandas as pd
from sqlalchemy import URL
# Local packages
from drapi import __version__ as drapiVersion
from drapi import loggingChoices
from drapi.code.drapi.drapi import (choosePathToLog,
                                    getTimestamp,
                                    loggingChoiceParser,
                                    makeDirPath)
# Super-local imports
from drapi.code.drapi.getData.getData import getData


if __name__ == "__main__":
    # >>> `Argparse` arguments >>>
    parser = argparse.ArgumentParser()

    # Arguments: Main
    parser.add_argument("--LEFT_DEIDENTIFICATION_MAP_PATH",
                        type=str,
                        required=True)
    parser.add_argument("--LEFT_VARIABLE_NAME",
                        type=str,
                        help="The column header for the left map.")
    parser.add_argument("--RIGHT_DEIDENTIFICATION_MAP_PATH",
                        type=str,
                        required=True)
    parser.add_argument("--RIGHT_VARIABLE_NAME",
                        type=str,
                        help="The column header for the right map.")
    parser.add_argument("--SQL_FILE_PATH",
                        type=Path,
                        help="The path to the SQL file to use for creating the left-to-right map.")
    parser.add_argument("--SQL_FILE_PLACEHOLDER",
                        type=str,
                        help="The SQL file template placeholder.")
    parser.add_argument("--OUTPUT_FILE_NAME",
                        type=str,
                        required=True,
                        help="The name of the query results.")
    parser.add_argument("--LEFT_VARIABLE_DEIDENTIFIED_NAME",
                        type=str,
                        help="The de-identified version of the left variable's name")
    parser.add_argument("--RIGHT_VARIABLE_DEIDENTIFIED_NAME",
                        type=str,
                        help="The de-identified version of the right variable's name")

    # Variable aliases
    parser.add_argument("--LEFT_VARIABLE_NAME_ALIAS_DEIDENTIFICATION_MAP",
                        type=str,
                        help="")
    parser.add_argument("--LEFT_VARIABLE_NAME_ALIAS_LEFT_TO_RIGHT_MAP",
                        type=str,
                        help="")
    parser.add_argument("--RIGHT_VARIABLE_NAME_ALIAS_DEIDENTIFICATION_MAP",
                        type=str,
                        help="")
    parser.add_argument("--RIGHT_VARIABLE_NAME_ALIAS_LEFT_TO_RIGHT_MAP",
                        type=str,
                        help="")
    parser.add_argument("--STANDARDIZE_COLUMN_NAMES",
                        action="store_true",
                        help="Whether to rename variable aliases to standard variable names.")

    parser.add_argument("--DROP_NA_HOW",
                        default="all",
                        choices=["all",
                                 "any"],
                        help="How to drop NAs from the final results.")

    parser.add_argument("--FIRST_TIME",
                        action="store_true")
    parser.add_argument("--OLD_RUN_PATH",
                        type=str,
                        help="""The path to the data previously downloaded from another session of this program. This is usually the "Left-to-right Map" directory.""")

    # Arguments: Meta-parameters
    parser.add_argument("--LOG_LEVEL",
                        default=10,
                        type=loggingChoiceParser,
                        choices=loggingChoices,
                        help="""Increase output verbosity. See "logging" module's log level for valid values.""")

    # Arguments: SQL connection settings
    parser.add_argument("--SERVER",
                        default="DWSRSRCH01.shands.ufl.edu",
                        type=str,
                        choices=["Acuo03.shands.ufl.edu",
                                 "EDW.shands.ufl.edu",
                                 "DWSRSRCH01.shands.ufl.edu",
                                 "IDR01.shands.ufl.edu",
                                 "RDW.shands.ufl.edu"],
                        help="")
    parser.add_argument("--DATABASE",
                        default="DWS_PROD",
                        type=str,
                        choices=["DWS_NOTES",
                                 "DWS_OMOP_PROD",
                                 "DWS_OMOP",
                                 "DWS_PROD"],  # TODO Add the i2b2 databases... or all the other databases?
                        help="")
    parser.add_argument("--USER_DOMAIN",
                        default="UFAD",
                        type=str,
                        choices=["UFAD"],
                        help="")
    parser.add_argument("--USERNAME",
                        default=os.environ["USER"],
                        type=str,
                        help="")
    parser.add_argument("--USER_ID",
                        default=None,
                        help="")
    parser.add_argument("--USER_PWD",
                        default=None,
                        help="")

    argNamespace = parser.parse_args()

    # Parsed arguments: Main
    LEFT_DEIDENTIFICATION_MAP_PATH = argNamespace.LEFT_DEIDENTIFICATION_MAP_PATH
    LEFT_VARIABLE_NAME = argNamespace.LEFT_VARIABLE_NAME
    RIGHT_DEIDENTIFICATION_MAP_PATH = argNamespace.RIGHT_DEIDENTIFICATION_MAP_PATH
    RIGHT_VARIABLE_NAME = argNamespace.RIGHT_VARIABLE_NAME
    SQL_FILE_PATH = argNamespace.SQL_FILE_PATH
    SQL_FILE_PLACEHOLDER = argNamespace.SQL_FILE_PLACEHOLDER
    OUTPUT_FILE_NAME = argNamespace.OUTPUT_FILE_NAME

    LEFT_VARIABLE_DEIDENTIFIED_NAME = argNamespace.LEFT_VARIABLE_DEIDENTIFIED_NAME
    RIGHT_VARIABLE_DEIDENTIFIED_NAME = argNamespace.RIGHT_VARIABLE_DEIDENTIFIED_NAME

    LEFT_VARIABLE_NAME_ALIAS_DEIDENTIFICATION_MAP = argNamespace.LEFT_VARIABLE_NAME_ALIAS_DEIDENTIFICATION_MAP
    LEFT_VARIABLE_NAME_ALIAS_LEFT_TO_RIGHT_MAP = argNamespace.LEFT_VARIABLE_NAME_ALIAS_LEFT_TO_RIGHT_MAP
    RIGHT_VARIABLE_NAME_ALIAS_DEIDENTIFICATION_MAP = argNamespace.RIGHT_VARIABLE_NAME_ALIAS_DEIDENTIFICATION_MAP
    RIGHT_VARIABLE_NAME_ALIAS_LEFT_TO_RIGHT_MAP = argNamespace.RIGHT_VARIABLE_NAME_ALIAS_LEFT_TO_RIGHT_MAP
    STANDARDIZE_COLUMN_NAMES = argNamespace.STANDARDIZE_COLUMN_NAMES

    DROP_NA_HOW = argNamespace.DROP_NA_HOW

    FIRST_TIME = argNamespace.FIRST_TIME
    OLD_RUN_PATH = argNamespace.OLD_RUN_PATH

    # Parsed arguments: Meta-parameters
    LOG_LEVEL = argNamespace.LOG_LEVEL

    # Parsed arguments: SQL connection settings
    SERVER = argNamespace.SERVER
    DATABASE = argNamespace.DATABASE
    USER_DOMAIN = argNamespace.USER_DOMAIN
    USERNAME = argNamespace.USERNAME
    USER_ID = argNamespace.USER_ID
    USER_PWD = argNamespace.USER_PWD
    # <<< `Argparse` arguments <<<

    # >>> Custom parsing >>>
    # If `STANDARDIZE_COLUMN_NAMES` is passed, we need `LEFT_VARIABLE_NAME` and `RIGHT_VARIABLE_NAME`
    if STANDARDIZE_COLUMN_NAMES:
        if isinstance(LEFT_VARIABLE_NAME, type(None)) or isinstance(LEFT_VARIABLE_NAME, type(None)):
            message = "If `STANDARDIZE_COLUMN_NAMES` is passed, we need `LEFT_VARIABLE_NAME` and `RIGHT_VARIABLE_NAME`"
            parser.error(message)
    # If `LEFT_VARIABLE_DEIDENTIFIED_NAME` is NOT passed, we need `LEFT_VARIABLE_NAME`
    if isinstance(LEFT_VARIABLE_DEIDENTIFIED_NAME, type(None)):
        if isinstance(LEFT_VARIABLE_NAME, type(None)):
            message = "If `LEFT_VARIABLE_DEIDENTIFIED_NAME` is NOT passed, we need `LEFT_VARIABLE_NAME`"
            parser.error(message)
    # If `RIGHT_VARIABLE_DEIDENTIFIED_NAME` is not passed, we need `RIGHT_VARIABLE_NAME`
    if isinstance(RIGHT_VARIABLE_DEIDENTIFIED_NAME, type(None)):
        if isinstance(RIGHT_VARIABLE_NAME, type(None)):
            message = "If `RIGHT_VARIABLE_DEIDENTIFIED_NAME` is not passed, we need `RIGHT_VARIABLE_NAME`"
            parser.error(message)
    # <<< Custom parsing <<<

    # >>> Argument checks >>>
    # NOTE TODO Look into handling this natively with `argparse` by using `subcommands`. See "https://stackoverflow.com/questions/30457162/argparse-with-different-modes"
    if FIRST_TIME and OLD_RUN_PATH:
        message = "It is ambiguous if you provide both `FIRST_TIME` and `OLD_RUN_PATH`. Please only choose one."
        parser.error(message=message)
    # <<< Argument checks <<<

    # Variables: Path construction: General
    runTimestamp = getTimestamp()
    thisFilePath = Path(__file__)
    thisFileStem = thisFilePath.stem
    currentWorkingDir = Path(os.getcwd()).absolute()
    projectDir = currentWorkingDir
    dataDir = projectDir.joinpath("data")
    if dataDir:
        inputDataDir = dataDir.joinpath("input")
        intermediateDataDir = dataDir.joinpath("intermediate")
        outputDataDir = dataDir.joinpath("output")
        if intermediateDataDir:
            runIntermediateDir = intermediateDataDir.joinpath(thisFileStem, runTimestamp)
        if outputDataDir:
            runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
    logsDir = projectDir.joinpath("logs")
    if logsDir:
        runLogsDir = logsDir.joinpath(thisFileStem)
    sqlDir = projectDir.joinpath("sql")

    # Variables: Path construction: Project-specific
    pass

    # Variables: SQL Parameters
    if USER_ID:
        userID = USER_ID[:]
    else:
        userID = fr"{USER_DOMAIN}\{USERNAME}"
    if USER_PWD:
        userPwd = USER_PWD
    else:
        userPwd = os.environ["HFA_UFADPWD"]
    conStr = f"mssql+pymssql://{userID}:{userPwd}@{SERVER}/{DATABASE}"

    # Variables: Other
    pass

    # Directory creation: General
    makeDirPath(runIntermediateDir)
    makeDirPath(runOutputDir)
    makeDirPath(runLogsDir)

    # Logging block
    logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
    logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

    logger = logging.getLogger(__name__)

    fileHandler = logging.FileHandler(logpath)
    fileHandler.setLevel(9)
    fileHandler.setFormatter(logFormat)

    streamHandler = logging.StreamHandler()
    streamHandler.setLevel(LOG_LEVEL)
    streamHandler.setFormatter(logFormat)

    logger.addHandler(fileHandler)
    logger.addHandler(streamHandler)

    logger.setLevel(9)

    logger.info(f"""Begin running "{choosePathToLog(path=thisFilePath, rootPath=projectDir)}".""")
    logger.info(f"""DRAPI-Lemur version is "{drapiVersion}".""")
    logger.info(f"""All other paths will be reported in debugging relative to the current working directory: "{choosePathToLog(path=projectDir, rootPath=projectDir)}".""")

    argList = argNamespace._get_args() + argNamespace._get_kwargs()
    argListString = pprint.pformat(argList)  # TODO Remove secrets from list to print, e.g., passwords.
    logger.info(f"""Script arguments:\n{argListString}""")

    # >>> Begin module body >>>

    # Map 1: Left de-identification map
    leftdf = pd.read_csv(LEFT_DEIDENTIFICATION_MAP_PATH)

    # Map 2: Left-to-right map
    if FIRST_TIME:
        left2rightMapDir = runOutputDir.joinpath("Left-to-right Map")
        left2rightMapDir.mkdir()
        connectionString = URL.create(drivername="mssql+pymssql",
                                      username=userID,
                                      password=userPwd,
                                      host=SERVER,
                                      database=DATABASE)

        getData(sqlFilePath=SQL_FILE_PATH,
                connectionString=connectionString,
                filterVariableChunkSize=10000,
                filterVariableColumnName=LEFT_VARIABLE_NAME_ALIAS_DEIDENTIFICATION_MAP,
                filterVariableData=leftdf,
                filterVariableFilePath=None,
                filterVariablePythonDataType="int",
                filterVariableSqlQueryTemplatePlaceholder=SQL_FILE_PLACEHOLDER,
                logger=logger,
                outputFileName=OUTPUT_FILE_NAME,
                runOutputDir=left2rightMapDir,
                queryChunkSize=10000)
    else:
        left2rightMapDir = Path(OLD_RUN_PATH)

    # Join maps: Concatenate left-to-right map
    listofPaths = sorted([fpath for fpath in left2rightMapDir.iterdir() if fpath.suffix.lower() == ".csv"])
    left2rightdf = pd.DataFrame()
    itTotal = len(listofPaths)
    for it, fpath in enumerate(listofPaths, start=1):
        logger.info(f"""  Working on file {it:,} of {itTotal:,}.""")
        df = pd.read_csv(fpath)
        left2rightdf = pd.concat([left2rightdf, df])
    del df

    # Map 3: Right de-identification map
    rightdf = pd.read_csv(RIGHT_DEIDENTIFICATION_MAP_PATH)

    # Join maps: Homogenize variable names
    pass

    # Join maps
    joinedMaps = leftdf.set_index(LEFT_VARIABLE_NAME_ALIAS_DEIDENTIFICATION_MAP).join(other=left2rightdf.set_index(LEFT_VARIABLE_NAME_ALIAS_LEFT_TO_RIGHT_MAP),
                                                                                      how="outer")
    joinedMaps = joinedMaps.reset_index(names=LEFT_VARIABLE_NAME_ALIAS_DEIDENTIFICATION_MAP)

    joinedMaps = joinedMaps.set_index(RIGHT_VARIABLE_NAME_ALIAS_LEFT_TO_RIGHT_MAP).join(other=rightdf.set_index(RIGHT_VARIABLE_NAME_ALIAS_LEFT_TO_RIGHT_MAP),
                                                                                        how="outer")
    joinedMaps = joinedMaps.reset_index(names=RIGHT_VARIABLE_NAME_ALIAS_LEFT_TO_RIGHT_MAP)

    # Standardize column names
    if STANDARDIZE_COLUMN_NAMES:
        joinedMaps = joinedMaps.rename(columns={LEFT_VARIABLE_NAME_ALIAS_DEIDENTIFICATION_MAP: f"{LEFT_VARIABLE_NAME}_L",
                                                LEFT_VARIABLE_NAME_ALIAS_LEFT_TO_RIGHT_MAP: f"{LEFT_VARIABLE_NAME}_LR",
                                                RIGHT_VARIABLE_NAME_ALIAS_DEIDENTIFICATION_MAP: f"{RIGHT_VARIABLE_NAME}_R",
                                                RIGHT_VARIABLE_NAME_ALIAS_LEFT_TO_RIGHT_MAP: f"{RIGHT_VARIABLE_NAME}_LR",
                                                LEFT_VARIABLE_DEIDENTIFIED_NAME: f"De-identified {LEFT_VARIABLE_NAME}",
                                                RIGHT_VARIABLE_DEIDENTIFIED_NAME: f"De-identified {RIGHT_VARIABLE_NAME}",
                                                # For the below options we are assuming the de-identification prefix is "De-identififed" and not "deid" or another variant. These lines can possibly be removed.
                                                f"De-identified {LEFT_VARIABLE_NAME_ALIAS_DEIDENTIFICATION_MAP}": f"De-identified {LEFT_VARIABLE_NAME}",
                                                f"De-identified {LEFT_VARIABLE_NAME_ALIAS_LEFT_TO_RIGHT_MAP}": f"De-identified {LEFT_VARIABLE_NAME}",
                                                f"De-identified {RIGHT_VARIABLE_NAME_ALIAS_DEIDENTIFICATION_MAP}": f"De-identified {RIGHT_VARIABLE_NAME}",
                                                f"De-identified {RIGHT_VARIABLE_NAME_ALIAS_LEFT_TO_RIGHT_MAP}": f"De-identified {RIGHT_VARIABLE_NAME}"})
    else:
        pass

    # Select data to output: Select column names
    # We are assuming that if the left and right de-identified variable names are not supplied we are using the standard variable names.
    if LEFT_VARIABLE_DEIDENTIFIED_NAME:
        leftDeidentifiedColumnName = LEFT_VARIABLE_DEIDENTIFIED_NAME
    else:
        leftDeidentifiedColumnName = f"De-identified {LEFT_VARIABLE_NAME}"
    if RIGHT_VARIABLE_DEIDENTIFIED_NAME:
        rightDeidentifiedColumnName = RIGHT_VARIABLE_DEIDENTIFIED_NAME
    else:
        rightDeidentifiedColumnName = f"De-identified {RIGHT_VARIABLE_NAME}"
    COLUMNS_TO_EXPORT = [leftDeidentifiedColumnName,
                         rightDeidentifiedColumnName]

    # Select data to output: Select columns
    logger.info(joinedMaps.columns)
    logger.info(joinedMaps.head().T)
    finalMap = joinedMaps[COLUMNS_TO_EXPORT]

    mapSize0 = finalMap.shape[0]
    finalMap = finalMap.dropna(how=DROP_NA_HOW)
    finalMap = finalMap.sort_values(by=COLUMNS_TO_EXPORT)
    exportPath = runOutputDir.joinpath(f"{LEFT_VARIABLE_NAME} to {RIGHT_VARIABLE_NAME}.CSV")
    finalMap.to_csv(exportPath, index=False)

    # QA
    mapSize1 = finalMap.shape[0]
    logger.info(f"""Final map shape before and after dropping "{DROP_NA_HOW}" NAs: {mapSize0:,} -> {mapSize1:,}.""")

    # Output location summary
    logger.info(f"""Results are in "{choosePathToLog(path=runOutputDir, rootPath=projectDir)}".""")

    # Remove intermediate files, unless running in `DEBUG` mode.
    if logger.getEffectiveLevel() > 10:
        logger.info("Removing intermediate files.")
        shutil.rmtree(runIntermediateDir)
        logger.info("Removing intermediate files - done.")

    # End module body
    logger.info(f"""Finished running "{choosePathToLog(path=thisFilePath, rootPath=projectDir)}".""")



"""
Get "Person ID" from "Patient Key" using `getData`.
"""

import logging
import os
import shutil
from pathlib import Path
# Third-party packages
import pandas as pd
# Local packages
from drapi.code.drapi.drapi import (getTimestamp,
                                    makeDirPath,
                                    successiveParents)
from drapi.code.drapi.getData.getData import getData

# Arguments
PATIENT_KEYS_CSV_FILE_PATH = ""  # TODO
PATIENT_KEYS_CSV_FILE_HEADER = ""  # TODO

# Arguments: Meta-variables
PROJECT_DIR_DEPTH = 2
DATA_REQUEST_DIR_DEPTH = PROJECT_DIR_DEPTH + 2
IRB_DIR_DEPTH = PROJECT_DIR_DEPTH + 2
IDR_DATA_REQUEST_DIR_DEPTH = IRB_DIR_DEPTH + 3

ROOT_DIRECTORY = "IRB_DIRECTORY"  # TODO One of the following:
                                                 # ["IDR_DATA_REQUEST_DIRECTORY",    # noqa
                                                 #  "IRB_DIRECTORY",                 # noqa
                                                 #  "DATA_REQUEST_DIRECTORY",        # noqa
                                                 #  "PROJECT_OR_PORTION_DIRECTORY"]  # noqa

LOG_LEVEL = "INFO"

# Arguments: SQL connection settings
SERVER = "DWSRSRCH01.shands.ufl.edu"
DATABASE = "DWS_OMOP_PROD"
USERDOMAIN = "UFAD"
USERNAME = os.environ["USER"]
UID = None
PWD = os.environ["HFA_UFADPWD"]

# Variables: Path construction: General
runTimestamp = getTimestamp()
thisFilePath = Path(__file__)
thisFileStem = thisFilePath.stem
projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_DIR_DEPTH)
IRBDir, _ = successiveParents(thisFilePath, IRB_DIR_DEPTH)
IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
dataDir = projectDir.joinpath("data")
if dataDir:
    inputDataDir = dataDir.joinpath("input")
    intermediateDataDir = dataDir.joinpath("intermediate")
    outputDataDir = dataDir.joinpath("output")
    if intermediateDataDir:
        runIntermediateDataDir = intermediateDataDir.joinpath(thisFileStem, runTimestamp)
    if outputDataDir:
        runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
logsDir = projectDir.joinpath("logs")
if logsDir:
    runLogsDir = logsDir.joinpath(thisFileStem)
sqlDir = projectDir.joinpath("sql")

if ROOT_DIRECTORY == "PROJECT_OR_PORTION_DIRECTORY":
    rootDirectory = projectDir
elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
    rootDirectory = dataRequestDir
elif ROOT_DIRECTORY == "IRB_DIRECTORY":
    rootDirectory = IRBDir
elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
    rootDirectory = IDRDataRequestDir
else:
    raise Exception("An unexpected error occurred.")

# Variables: Path construction: Project-specific
personID_SQLQueryFilePath = sqlDir.joinpath("grabPersonIDs.SQL")

# Variables: SQL Parameters
if UID:
    uid = UID[:]
else:
    uid = fr"{USERDOMAIN}\{USERNAME}"
conStr = f"mssql+pymssql://{uid}:{PWD}@{SERVER}/{DATABASE}"

# Variables: Other
pass

# Directory creation: General
makeDirPath(runIntermediateDataDir)
makeDirPath(runOutputDir)
makeDirPath(runLogsDir)

# Directory creation: Project-specific
pass

# Logging block
logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

logger = logging.getLogger(__name__)

fileHandler = logging.FileHandler(logpath)
fileHandler.setLevel(9)
fileHandler.setFormatter(logFormat)

streamHandler = logging.StreamHandler()
streamHandler.setLevel(LOG_LEVEL)
streamHandler.setFormatter(logFormat)

logger.addHandler(fileHandler)
logger.addHandler(streamHandler)

logger.setLevel(9)

if __name__ == "__main__":
    logger.info(f"""Begin running "{thisFilePath}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Script arguments:

    # Arguments
    `PATIENT_KEYS_CSV_FILE_PATH`: "{PATIENT_KEYS_CSV_FILE_PATH}"
    `PATIENT_KEYS_CSV_FILE_HEADER`: "{PATIENT_KEYS_CSV_FILE_HEADER}"

    # Arguments: Meta-arguments
    `PROJECT_DIR_DEPTH`: "{PROJECT_DIR_DEPTH}" ----------> "{projectDir}"
    `IRB_DIR_DEPTH`: "{IRB_DIR_DEPTH}" --------------> "{IRBDir}"
    `IDR_DATA_REQUEST_DIR_DEPTH`: "{IDR_DATA_REQUEST_DIR_DEPTH}" -> "{IDRDataRequestDir}"

    `LOG_LEVEL` = "{LOG_LEVEL}"

    # Arguments: SQL connection settings
    `SERVER` = "{SERVER}"
    `DATABASE` = "{DATABASE}"
    `USERDOMAIN` = "{USERDOMAIN}"
    `USERNAME` = "{USERNAME}"
    `UID` = "{UID}"
    `PWD` = censored
    """)

    # Script
    # Get patient keys
    data = pd.read_csv(PATIENT_KEYS_CSV_FILE_PATH)

    # Get input file header
    nColumns = data.shape[1]
    if PATIENT_KEYS_CSV_FILE_HEADER:
        inputFileHeader = PATIENT_KEYS_CSV_FILE_HEADER
    else:
        if nColumns == 1:
            inputFileHeader = data.columns[0]
        else:
            message = f"""The input data has more than one column and you did not specify the column to use. Please set a value for `PATIENT_KEYS_CSV_FILE_HEADER`."""
            logger.critical(message)
            raise Exception(message)

    patientKeysInput = data[inputFileHeader].astype(int).drop_duplicates().sort_values()

    # Query person IDs
    logger.info("""Querying database for person IDs.""")
    queryResultsDir = runIntermediateDataDir.joinpath("Query Results")
    queryResultsDir.mkdir()
    getData(sqlFilePath=personID_SQLQueryFilePath,
            connectionString=conStr,
            filterVariablePythonDataType="int",
            filterVariableSqlQueryTemplatePlaceholder="{PYTHON_VARIABLE: Patient Key}",
            logger=logger,
            outputFileName="Person ID",
            runOutputDir=queryResultsDir,
            filterVariableColumnName=PATIENT_KEYS_CSV_FILE_HEADER,
            filterVariableFilePath=PATIENT_KEYS_CSV_FILE_PATH)
    logger.info("""Querying database for person IDs - done.""")

    # Load query results
    queryResults = pd.DataFrame()
    resultsList = sorted(queryResultsDir.iterdir())
    it1Total = len(resultsList)
    for it1, fpath in enumerate(resultsList, start=1):
        logger.info(f"""  Loading query result {it1:,} of {it1Total:,}.""")
        df = pd.read_csv(fpath)
        queryResults = pd.concat([queryResults, df])

    # Compare number of Person IDs returned with Patient Keys queried
    patientKeysInput = pd.DataFrame(patientKeysInput)
    patientKeysInput = patientKeysInput.rename(columns={inputFileHeader: "Patient Key"})
    patientKeysInput["Patient Key"] = patientKeysInput["Patient Key"].astype(int)
    queryResults["Patient Key"] = queryResults["Patient Key"].astype(int)
    patientKeysInput["Found"] = patientKeysInput["Patient Key"].isin(queryResults["Patient Key"])
    personIDsFound = patientKeysInput.set_index("Patient Key").join(queryResults.set_index("Patient Key"), "Patient Key", "left")
    personIDsFound = personIDsFound.sort_values("person_id")

    # Summary statistics
    numFound = personIDsFound["Found"].sum()
    numPatientKeys = len(patientKeysInput)
    logger.info(f"""A total of {numFound:,} patient keys of {numPatientKeys:,} were mapped to OMOP person IDs.""")
    logger.info("""If the number of person IDs is LESS THAN the number of patient keys, it's possible that some patients were merged BEFORE the OMOP database was updated.""")
    logger.info("""If the number of person IDs is GREATER THAN the number of patient keys, it's possible that some patients were merged AFTER the OMOP database was updated.""")

    # If any column contains NaNs, convert the column to the "Object" data type, to preserve data quality
    personIDsFound2 = personIDsFound.copy()
    for column in personIDsFound.columns:
        if personIDsFound[column].isna().sum() > 0:
            li = []
            for value in personIDsFound[column].values:
                if pd.isna(value):
                    li.append("")
                else:
                    li.append(str(int(value)))
            personIDsFound2[column] = li

    # Save results: Person IDs found
    personIDsFoundExportPath = runOutputDir.joinpath("Person IDs - QA.csv")
    personIDsFound2.to_csv(personIDsFoundExportPath)
    logger.info(f"""The map of patient keys to person IDs, and those missing or found, was saved to "{personIDsFoundExportPath.absolute().relative_to(rootDirectory)}".""")

    # Save results: person IDs
    personIDs = personIDsFound["person_id"].drop_duplicates().dropna().sort_values().astype(int)
    personIDsExportPath = runOutputDir.joinpath("Person IDs.csv")
    personIDs.to_csv(personIDsExportPath, index=False)
    logger.info(f"""Person IDs were saved to "{personIDsExportPath.relative_to(projectDir)}".""")

    # Remove intermediate files
    shutil.rmtree(runIntermediateDataDir)

    # End script
    logger.info(f"""Finished running "{thisFilePath.relative_to(projectDir)}".""")



"""
A collection of classes
"""

class SecretString(str):
    """A `str` obj with a special `__repr__` that doesn't reveal the value."""

    def __init__(self, object=""):
        str.__init__(self)

    def __repr__(self):
        return "CENSORED"

    def __str__(self):
        return "CENSORED"



"""
Template script for uploading pandas-compatible files to a SQL server - functionless implementation.
"""

import os
from pathlib import Path

import pandas as pd
import sqlalchemy as sa

from drapi.code.drapi.drapi import (getTimestamp,
                                    readDataFile)

DIRECTORY_OR_LIST_OF_PATHS = "../../Data Request 03 - 2024-04-15/Intermediate Results/Clinical Text Portion/data/output/freeText/2024-06-04 17-14-41/free_text/Sepsis_order_impression"
TABLE_NAME = f"DR IRB202202722 Text - OI {getTimestamp()}"
SCHEMA_NAME = "dbo"
CONNECTION_STRING = f"""mssql+pymssql://UFAD\herman:{os.environ["HFA_UFADPWD"]}@DWSRSRCH01.shands.ufl.edu/DWS_OMOP"""

if Path(DIRECTORY_OR_LIST_OF_PATHS).is_dir():
    list_of_paths = sorted(list(Path(DIRECTORY_OR_LIST_OF_PATHS).iterdir()))
elif isinstance(DIRECTORY_OR_LIST_OF_PATHS, list):
    list_of_paths = sorted([Path(el) for el in DIRECTORY_OR_LIST_OF_PATHS])
else:
    raise Exception(f"""`DIRECTORY_OR_LIST_OF_PATHS` should be a directory or a list of file paths.""")

print("Dropping table if it exists.")
query_prep_table = f"""\
DROP TABLE IF EXISTS [{SCHEMA_NAME}].[{TABLE_NAME}]
\
"""
with sa.engine.create_engine(url=CONNECTION_STRING).connect() as connection:
    with connection.begin():
        _ = connection.execute(sa.text(query_prep_table))
print("Dropping table if it exists - done.")

num_files = len(list_of_paths)
for it, fpath in enumerate(list_of_paths, start=1):
    print(f"""  Working on file {it:,} of {num_files:,}.""")
    df = readDataFile(fname=fpath,
                      engine="pyarrow")
    df = pd.DataFrame(df)  # For type hinting
    df.to_sql(name=TABLE_NAME,
              con=CONNECTION_STRING,
              schema=SCHEMA_NAME,
              if_exists="append",
              index=False)
    print(f"""  Working on file {it:,} of {num_files:,} - done.""")



"""
Uploads a pandas-compatible file to a SQL server
"""

import logging
import re
from pathlib import Path
from typing_extensions import (List,
                               Union)
# Third-party packages
import pandas as pd
import sqlalchemy as sa
# Local packages
from drapi.code.drapi.classes import SecretString
from drapi.code.drapi.drapi import readDataFile

def lazy_hack_1_function(string_: str):
    """
    """
    if isinstance(string_, str):
        new_string = re.sub(pattern="\x00",
                        repl="<NULL-BYTE>",
                        string=string_)
    else:
        new_string = string_
    return new_string


def uploadData(list_of_paths: List[Union[Path, str]],
               schema_name: str,
               table_name: str,
               connection_string: SecretString,
               lazy_hack_1: List[Union[int, str]],
               logger: logging.Logger):
    """
    Uploads a pandas-compatible file to a SQL server
    """
    logger.info("Dropping table if it exists.")
    query_prep_table = f"""\
    DROP TABLE IF EXISTS [{schema_name}].[{table_name}]
    \
    """
    with sa.engine.create_engine(url=connection_string).connect() as connection:
        with connection.begin():
            _ = connection.execute(sa.text(query_prep_table))
    logger.info("Dropping table if it exists - done.")

    num_files = len(list_of_paths)
    for it, fpath in enumerate(list_of_paths, start=1):
        logger.info(f"""  Working on file {it:,} of {num_files:,}.""")
        df = readDataFile(fname=fpath,
                          engine="pyarrow")
        df = pd.DataFrame(df)  # For type hinting
        if lazy_hack_1:
            column_name = "note_text"
            series = df[column_name]
            df[column_name] = series.apply(lazy_hack_1_function)
        df.to_sql(name=table_name,
                con=connection_string,
                schema=schema_name,
                if_exists="append",
                index=False)
        logger.info(f"""  Working on file {it:,} of {num_files:,} - done.""")



"""
Parsers used by argparse.
"""

import ast


def parse_string_to_boolean(string_: str) -> bool:
    """
    """
    if string_.lower() == "true":
        return True
    elif string_.lower() == "false":
        return False
    else:
        raise Exception("String must be one of {true, false}, case insensitive.")


def parse_string_to_bytes(string_: str) -> bytes:
    """
    Expects `string_` to be a Python string literal representation of a byte like `"\\\\x00"` or `"a"`.
    """
    # lazy hack 1
    if string_ == "'":
        string_1 = f'b"{string_}"'
    elif string_ == '"':
        string_1 = f"b'{string_}'"
    else:
        string_1 = f"""b'{string_}'"""
    bytes_array = ast.literal_eval(string_1)
    return bytes_array



"""
An example of how to use the BO/SQL features of DRAPI-Lemur
"""

import logging
# Third-party packages
import pandas as pd
# First-party packages
from drapi.code.drapi.getData.getData import getData

getData(sqlFilePath="Query.SQL",
        connectionString="mssql+pymssql://USERNAME:PASSWORD@SERVER/DATABASE",
        filterVariableChunkSize=10000,
        filterVariableColumnName="Column Name",
        filterVariableData=pd.Series(),
        filterVariableFilePath=None,
        filterVariablePythonDataType="int",
        filterVariableSqlQueryTemplatePlaceholder="Placeholder Text in SQL Query File",
        logger=logging.getLogger(),
        outputFileName="Query Result Name",
        runOutputDir="My Directory",
        queryChunkSize=10000)
